import sys
import numpy
import xraylib
from PyQt5.QtWidgets import QApplication, QMessageBox, QSizePolicy

from orangewidget import gui
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui, congruence
from oasys.widgets.exchange import DataExchangeObject
from xoppylib.power.xoppy_calc_power import xoppy_calc_power

from oasys.widgets.exchange import DataExchangeObject
from orangecontrib.xoppy.widgets.gui.ow_xoppy_widget import XoppyWidget


import scipy.constants as codata

class Scintillator(XoppyWidget):
    name = "Scintillator"
    id = "orange.widgets.dataxpower"
    description = "Power Absorbed and Transmitted by Optical Elements"
    icon = "icons/id19_scintillator.png"
    priority = 4
    category = ""
    keywords = ["xoppy", "power", "scintillator"]

    inputs = [("ExchangeData", DataExchangeObject, "acceptExchangeData")]

    SOURCE = Setting(2)
    SCINTILLATOR = Setting(2)
    THICK = Setting(500.0)
    #DENS = Setting('?')
    ENER_MIN = Setting(10000.0)
    ENER_MAX = Setting(200000.0)
    ENER_N = Setting(2000)
    SOURCE_FILE = Setting("?")
    CORRECTION_FACTOR = Setting(0.1)
    FILE_DUMP = 0

    def build_gui(self):

        self.leftWidgetPart.setSizePolicy(QSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding))
        self.leftWidgetPart.setMaximumWidth(self.CONTROL_AREA_WIDTH + 20)
        self.leftWidgetPart.updateGeometry()

        box = oasysgui.widgetBox(self.controlArea, self.name + " Input Parameters", orientation="vertical", width=self.CONTROL_AREA_WIDTH-10)

        idx = -1

        # widget index 1
        idx += 1
        box1 = gui.widgetBox(box)
        self.box_source = gui.comboBox(box1, self, "SOURCE",
                                       label=self.unitLabels()[idx], addSpace=False,
                                       items=['From Oasys wire', 'Normalized to 1',
                                              'From external file.                '],
                                       valueType=int, orientation="horizontal", labelWidth=150)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 2
        idx += 1
        box1 = gui.widgetBox(box)
        self.box_source = gui.comboBox(box1, self, "SCINTILLATOR",
                                       label=self.unitLabels()[idx], addSpace=False,
                                       items=['Diamond','Gadox','LuAg','GGG','LSO','YAG','CWO'],
                                       valueType=int, orientation="horizontal", labelWidth=150)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 3
        idx += 1
        box1 = gui.widgetBox(box)
        oasysgui.lineEdit(box1, self, "THICK",
                                       label=self.unitLabels()[idx], addSpace=False,
                                       valueType=float,  orientation="horizontal", labelWidth=150)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 4
       # idx += 1
        #box1 = gui.widgetBox(box)
        #oasysgui.lineEdit(box1, self, "DENS",
                          #label=self.unitLabels()[idx], addSpace=False,
                          #valueType=float, orientation="horizontal", labelWidth=150)
        #self.show_at(self.unitFlags()[idx], box1)

        # widget index 5
        idx += 1
        box1 = gui.widgetBox(box)
        oasysgui.lineEdit(box1, self, "ENER_MIN",
                          label=self.unitLabels()[idx], addSpace=False,
                          valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 6
        idx += 1
        box1 = gui.widgetBox(box)
        oasysgui.lineEdit(box1, self, "ENER_MAX",
                          label=self.unitLabels()[idx], addSpace=False,
                          valueType=float, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 7
        idx += 1
        box1 = gui.widgetBox(box)
        oasysgui.lineEdit(box1, self, "ENER_N",
                          label=self.unitLabels()[idx], addSpace=False,
                          valueType=int, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1)

        # widget index 8 ***********   File Browser ******************
        idx += 1
        box1 = gui.widgetBox(box)
        file_box_id = oasysgui.widgetBox(box1, "", addSpace=False, orientation="horizontal")
        self.file_id = oasysgui.lineEdit(file_box_id, self, "SOURCE_FILE", self.unitLabels()[idx],
                                         labelWidth=100, valueType=str, orientation="horizontal")
        gui.button(file_box_id, self, "...", callback=self.select_input_file, width=25)
        self.show_at(self.unitFlags()[idx], box1)


        # widget index 9
        idx += 1
        box1 = gui.widgetBox(box)
        oasysgui.lineEdit(box1, self, "CORRECTION_FACTOR",
                                       label=self.unitLabels()[idx], addSpace=False,
                                       valueType=float, orientation="horizontal", labelWidth=150)
        self.show_at(self.unitFlags()[idx], box1)

        #widget index 10
        idx += 1
        box1 = gui.widgetBox(box)
        gui.separator(box1, height=7)

        gui.comboBox(box1, self, "FILE_DUMP",
                     label=self.unitLabels()[idx], addSpace=False,
                    items=['No', 'Yes (Scinti.spec)'],
                    valueType=int, orientation="horizontal", labelWidth=250)
        self.show_at(self.unitFlags()[idx], box1)

        self.input_spectrum = None




    def select_input_file(self):
        self.file_id.setText(oasysgui.selectFileFromDialog(self, self.SOURCE_FILE,
                                    "Open 2-columns file with spectral power",
                                    file_extension_filter="ascii dat (*.dat *.txt *spec)"))



    def unitLabels(self):
         return ['Input beam:', 'Scintillator','Thickness [microns]',
                #Density g/cm^3,
                 'From energy [eV]:      ',
                 'To energy [eV]:',
                 'Energy points:  ',
                 'File with input beam spectral power:', 'Correction factor',"Dump file"]


    def unitFlags(self):
         return ['True','True','True',
                #True,
                 'self.SOURCE  ==  1',
                 'self.SOURCE  ==  1',
                 'self.SOURCE  ==  1',
                 'self.SOURCE  ==  2',
                 'True','True']

    def get_help_name(self):
        return 'Scintillator'

    def selectFile(self):
        self.le_source_file.setText(oasysgui.selectFileFromDialog(self, self.SOURCE_FILE, "Open Source File", file_extension_filter="*.*"))

    def acceptExchangeData(self, exchangeData):

        self.input_spectrum = None
        self.SOURCE = 0
        # self.box_source.setCurrentIndex(self.SOURCE)

        try:
            if not exchangeData is None:
                if exchangeData.get_program_name() == "XOPPY":
                    no_bandwidth = False
                    if exchangeData.get_widget_name() =="UNDULATOR_FLUX" :
                        # self.SOURCE_FILE = "xoppy_undulator_flux"
                        no_bandwidth = True
                        index_flux = 2
                    elif exchangeData.get_widget_name() == "BM" :
                        if exchangeData.get_content("is_log_plot") == 1:
                            raise Exception("Logaritmic X scale of Xoppy Energy distribution not supported")
                        if exchangeData.get_content("calculation_type") == 0 and exchangeData.get_content("psi") == 0:
                            # self.SOURCE_FILE = "xoppy_bm_flux"
                            no_bandwidth = True
                            index_flux = 6
                        else:
                            raise Exception("Xoppy result is not an Flux vs Energy distribution integrated in Psi")
                    elif exchangeData.get_widget_name() =="XWIGGLER" :
                        # self.SOURCE_FILE = "xoppy_xwiggler_flux"
                        no_bandwidth = True
                        index_flux = 2
                    elif exchangeData.get_widget_name() =="WS" :
                        # self.SOURCE_FILE = "xoppy_xwiggler_flux"
                        no_bandwidth = True
                        index_flux = 2
                    elif exchangeData.get_widget_name() =="XTUBES" :
                        # self.SOURCE_FILE = "xoppy_xtubes_flux"
                        index_flux = 1
                        no_bandwidth = True
                    elif exchangeData.get_widget_name() =="XTUBE_W" :
                        # self.SOURCE_FILE = "xoppy_xtube_w_flux"
                        index_flux = 1
                        no_bandwidth = True
                    elif exchangeData.get_widget_name() =="BLACK_BODY" :
                        # self.SOURCE_FILE = "xoppy_black_body_flux"
                        no_bandwidth = True
                        index_flux = 2

                    elif exchangeData.get_widget_name() =="UNDULATOR_RADIATION" :
                        # self.SOURCE_FILE = "xoppy_undulator_radiation"
                        no_bandwidth = True
                        index_flux = 1
                    elif exchangeData.get_widget_name() =="POWER" :
                        # self.SOURCE_FILE = "xoppy_undulator_power"
                        no_bandwidth = True
                        index_flux = -1
                    elif exchangeData.get_widget_name() =="POWER3D" :
                        # self.SOURCE_FILE = "xoppy_power3d"
                        no_bandwidth = True
                        index_flux = 1

                    else:
                        raise Exception("Xoppy Source not recognized")

                    # self.SOURCE_FILE += "_" + str(id(self)) + ".dat"


                    spectrum = exchangeData.get_content("xoppy_data")

                    if exchangeData.get_widget_name() =="UNDULATOR_RADIATION" or \
                        exchangeData.get_widget_name() =="POWER3D":
                        [p, e, h, v ] = spectrum
                        tmp = p.sum(axis=2).sum(axis=1)*(h[1]-h[0])*(v[1]-v[0])*codata.e*1e3
                        spectrum = numpy.vstack((e,p.sum(axis=2).sum(axis=1)*(h[1]-h[0])*(v[1]-v[0])*
                                                 codata.e*1e3))
                        self.input_spectrum = spectrum
                    else:

                        if not no_bandwidth:
                            spectrum[:,index_flux] /= 0.001*spectrum[:,0]

                        self.input_spectrum = numpy.vstack((spectrum[:,0],spectrum[:,index_flux]))

                    self.process_showers()
                    self.compute()

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                       str(exception),
                QMessageBox.Ok)

            #raise exception




    def check_fields(self):
        self.CORRECTION_FACTOR = congruence.checkPositiveNumber(self.CORRECTION_FACTOR, "Correction Factor")
        self.THICK = congruence.checkStrictlyPositiveNumber(self.THICK, "Thickness")
        #self.DENS = congruence.checkStrictlyPositiveNumber(self.DENS, "Density")
        if self.SOURCE == 1:
            self.ENER_MIN = congruence.checkPositiveNumber(self.ENER_MIN, "Energy from")
            self.ENER_MAX = congruence.checkStrictlyPositiveNumber(self.ENER_MAX, "Energy to")
            congruence.checkLessThan(self.ENER_MIN, self.ENER_MAX, "Energy from", "Energy to")
            self.NPOINTS = congruence.checkStrictlyPositiveNumber(self.ENER_N, "Energy Points")
        elif self.SOURCE == 2:
            congruence.checkFile(self.SOURCE_FILE)

    def do_xoppy_calculation(self):
        return self.xoppy_calc_xpower()

    def extract_data_from_xoppy_output(self, calculation_output):
        return calculation_output


    def get_data_exchange_widget_name(self):
        return "POWER"




    def getTitles(self):
        return ['Input Beam','Total CS','Mu','Transmitivity','Absorption after correction','Intensity','Intensity after correction']

    def getXTitles(self):
        return ["Energy [eV]","Energy [eV]","Energy [eV]","Energy [eV]","Energy [eV]","Energy [eV]", 'Energy [eV]']

    def getYTitles(self):
        return ["Source",'Total CS cm^2/g','Mu cm^-1','Transmitivity','Absorption',"Intensity", 'Intensity']


    def getVariablesToPlot(self):
        return [(0, 1),(0, 2),(0, 3),(0, 4),(0, 5),(0, 6),(0,7)]

    def getLogPlot(self):
        return [(False,False),(False, True),(False, True),(False, False),(False, False),(False,False),(False,False)]


    def formula(self):
        if self.SCINTILLATOR == 0:
            F='C'
        if self.SCINTILLATOR == 1:
            F='Gd2O2S'
        if self.SCINTILLATOR == 2:
            F='Al5Lu3O12'
        if self.SCINTILLATOR == 3:
            F='Gd3Ga5O12'
        if self.SCINTILLATOR == 4:
            F='Lu2SiO5'
        if self.SCINTILLATOR == 5:
            F='Y3Al5O12'
        if self.SCINTILLATOR == 6:
            F='CdWO4'
        return F

    def density(self):
        if self.SCINTILLATOR == 0:
            D=3.508
        if self.SCINTILLATOR == 1:
            D=7.32/2 #50% powder
        if self.SCINTILLATOR == 2:
            D=6.76
        if self.SCINTILLATOR == 3:
            D=7.08
        if self.SCINTILLATOR == 4:
            D=7.4
        if self.SCINTILLATOR == 5:
            D=4.55
        if self.SCINTILLATOR == 6:
            D=7.9
        return D

    def correction(self,E,F):
        L=[]
        if self.CORRECTION_FACTOR==0:
            return(F)
        else :
            for k in range(len(E)):
                L.append(E[k]/1000 *self.CORRECTION_FACTOR*F[k])
            return(L)


    def xoppy_calc_xpower(self):

        substance=[self.formula()]
        thick=[self.THICK*1e-3]
        dens=[self.density()]
        #dens=[self.DENS]
        flags=[0]



        if self.SOURCE == 0:
            if self.input_spectrum is None:
                raise Exception("No input beam")
            else:
                energies = self.input_spectrum[0,:].copy()
                source = self.input_spectrum[1,:].copy()
        elif self.SOURCE == 1:
            energies = numpy.linspace(self.ENER_MIN,self.ENER_MAX,self.ENER_N)
            source = numpy.ones(energies.size)
            tmp = numpy.vstack( (energies,source))
            self.input_spectrum = source
        elif self.SOURCE == 2:
            if self.SOURCE == 2: source_file = self.SOURCE_FILE
            try:
                tmp = numpy.loadtxt(source_file)
                energies = tmp[:,0]
                source = tmp[:,1]
                self.input_spectrum = source
            except:
                print("Error loading file %s "%(source_file))
                raise

        if self.FILE_DUMP == 0:
            output_file = None
        else:
            output_file = "Scinti.spec"
            

        out_dictionary = xoppy_calc_power(energies=energies, source=source, substance=substance,
                                     flags=flags, dens=dens, thick=thick, angle=[], roughness=[], material_constants_library=xraylib)

        L=[]
        for k in range (5):
            L.append((out_dictionary['data'][k]).tolist())
        L.append(self.correction(L[0],(out_dictionary['data'][5]).tolist()))
        L.append((out_dictionary['data'][6]).tolist())
        L.append(List_Product([L[5],L[1]]))
        out_dictionary['data']=numpy.array(L)

        try:
             (out_dictionary["info"])
        except:
            pass
        # send exchange
        calculated_data = DataExchangeObject("XOPPY", self.get_data_exchange_widget_name())

        try:
            calculated_data.add_content("xoppy_data", out_dictionary["data"].T)
            calculated_data.add_content("plot_x_col", 0)
            calculated_data.add_content("plot_y_col", -1)
        except:
            pass
        try:
            calculated_data.add_content("labels", out_dictionary["labels"])
        except:
            pass
        try:
            calculated_data.add_content("info", out_dictionary["info"])
        except:
            pass

        return calculated_data


def List_Product(list):
    L = []
    l = 1
    for k in range(len(list[0])):
        for i in range(len(list)):
            l = l * list[i][k]
        L.append(l)
        l = 1
    return (L)


if __name__ == "__main__":

    from oasys.widgets.exchange import DataExchangeObject

    input_data_type = "POWER"

    if input_data_type == "POWER":
        # create fake UNDULATOR_FLUX xoppy exchange data
        e = numpy.linspace(1000.0, 10000.0, 100)
        source = e/10
        received_data = DataExchangeObject("XOPPY", "POWER")
        received_data.add_content("xoppy_data", numpy.vstack((e,e,source)).T)
        received_data.add_content("xoppy_code", "US")

    elif input_data_type == "POWER3D":
        # create unulator_radiation xoppy exchange data
        from xoppylib.sources.xoppy_undulators import xoppy_calc_undulator_radiation

        e, h, v, p, code = xoppy_calc_undulator_radiation(ELECTRONENERGY=6.04,ELECTRONENERGYSPREAD=0.001,ELECTRONCURRENT=0.2,\
                                           ELECTRONBEAMSIZEH=0.000395,ELECTRONBEAMSIZEV=9.9e-06,\
                                           ELECTRONBEAMDIVERGENCEH=1.05e-05,ELECTRONBEAMDIVERGENCEV=3.9e-06,\
                                           PERIODID=0.018,NPERIODS=222,KV=1.68,DISTANCE=30.0,
                                           SETRESONANCE=0,HARMONICNUMBER=1,
                                           GAPH=0.001,GAPV=0.001,\
                                           HSLITPOINTS=41,VSLITPOINTS=41,METHOD=0,
                                           PHOTONENERGYMIN=7000,PHOTONENERGYMAX=8100,PHOTONENERGYPOINTS=20,
                                           USEEMITTANCES=1)
        received_data = DataExchangeObject("XOPPY", "POWER3D")
        received_data = DataExchangeObject("XOPPY", "UNDULATOR_RADIATION")
        received_data.add_content("xoppy_data", [p, e, h, v])
        received_data.add_content("xoppy_code", code)


    app = QApplication(sys.argv)
    w = Scintillator()
    w.acceptExchangeData(received_data)
    w.show()
    app.exec()
    w.saveSettings()



import sys
from orangewidget import gui
from PyQt5.QtWidgets import QApplication

from orangecontrib.shadow.widgets.gui import ow_conic_coefficients_element, ow_optical_element
from orangecontrib.shadow.util import ShadowOpticalElement

import copy
# from orangecontrib.shadow.util.shadow_objects import ConicCoefficientsPreProcessorData # TODO to be added
from orangecontrib.esrf.shadow.util.shadow_objects import ConicCoefficientsPreProcessorData

class ConicCoefficientsMirror(ow_conic_coefficients_element.ConicCoefficientsElement):

    name = "Conic Coefficients Mirror"
    description = "Shadow OE: Conic Coefficients Mirror"
    icon = "icons/conic_coefficients_mirror.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 8
    category = "Optical Elements"
    keywords = ["data", "file", "load", "read"]

    inputs = copy.deepcopy(ow_optical_element.OpticalElement.inputs)
    inputs.append(("ConicCoeff_PreProcessor Data", ConicCoefficientsPreProcessorData, "SetConicCoeffPreProcessorData"))


    def __init__(self):
        graphical_Options=ow_optical_element.GraphicalOptions(is_mirror=True)

        super().__init__(graphical_Options)

        gui.rubber(self.controlArea)

        gui.rubber(self.mainArea)

    ################################################################
    #
    #  SHADOW MANAGEMENT
    #
    ################################################################

    def instantiateShadowOE(self):
        return ShadowOpticalElement.create_conic_coefficients_mirror()

    def SetConicCoeffPreProcessorData(self, data):
        if data is not None:
            verbose = 0

            if verbose:
                print("\nReceived data....")
                print("\n********************Data before, new: ")

                print("title", data.title)
                print("conic_coefficient_0",      self.conic_coefficient_0, data.conic_coefficient_0)
                print("conic_coefficient_1",      self.conic_coefficient_1, data.conic_coefficient_1)
                print("conic_coefficient_2",      self.conic_coefficient_2, data.conic_coefficient_2)
                print("conic_coefficient_3",      self.conic_coefficient_3, data.conic_coefficient_3)
                print("conic_coefficient_4",      self.conic_coefficient_4, data.conic_coefficient_4)
                print("conic_coefficient_5",      self.conic_coefficient_5, data.conic_coefficient_5)
                print("conic_coefficient_6",      self.conic_coefficient_6, data.conic_coefficient_6)
                print("conic_coefficient_7",      self.conic_coefficient_7, data.conic_coefficient_7)
                print("conic_coefficient_8",      self.conic_coefficient_8, data.conic_coefficient_8)
                print("conic_coefficient_9",      self.conic_coefficient_9, data.conic_coefficient_9)
                print("source_plane_distance",    self.source_plane_distance, data.source_plane_distance)
                print("image_plane_distance",     self.image_plane_distance, data.image_plane_distance)
                print("angles_respect_to",        self.angles_respect_to, data.angles_respect_to)
                print("incidence_angle_deg",      self.incidence_angle_deg, data.incidence_angle_deg)
                print("reflection_angle_deg",     self.reflection_angle_deg, data.reflection_angle_deg)
                print("mirror_orientation_angle", self.mirror_orientation_angle, data.mirror_orientation_angle)

            self.conic_coefficient_0 = data.conic_coefficient_0
            self.conic_coefficient_1 = data.conic_coefficient_1
            self.conic_coefficient_2 = data.conic_coefficient_2
            self.conic_coefficient_3 = data.conic_coefficient_3
            self.conic_coefficient_4 = data.conic_coefficient_4
            self.conic_coefficient_5 = data.conic_coefficient_5
            self.conic_coefficient_6 = data.conic_coefficient_6
            self.conic_coefficient_7 = data.conic_coefficient_7
            self.conic_coefficient_8 = data.conic_coefficient_8
            self.conic_coefficient_9 = data.conic_coefficient_9

            if data.source_plane_distance    is not None: self.source_plane_distance    = data.source_plane_distance
            if data.image_plane_distance     is not None: self.image_plane_distance     = data.image_plane_distance
            if data.angles_respect_to        is not None: self.angles_respect_to        = data.angles_respect_to
            if data.incidence_angle_deg      is not None: self.incidence_angle_deg      = data.incidence_angle_deg
            if data.reflection_angle_deg     is not None: self.reflection_angle_deg     = data.reflection_angle_deg
            if data.mirror_orientation_angle is not None: self.mirror_orientation_angle = data.mirror_orientation_angle

            if verbose:
                print("\n********************Data after: ")

                print("conic_coefficient_0", self.conic_coefficient_0)
                print("conic_coefficient_1", self.conic_coefficient_1)
                print("conic_coefficient_2", self.conic_coefficient_2)
                print("conic_coefficient_3", self.conic_coefficient_3)
                print("conic_coefficient_4", self.conic_coefficient_4)
                print("conic_coefficient_5", self.conic_coefficient_5)
                print("conic_coefficient_6", self.conic_coefficient_6)
                print("conic_coefficient_7", self.conic_coefficient_7)
                print("conic_coefficient_8", self.conic_coefficient_8)
                print("conic_coefficient_9", self.conic_coefficient_9)
                print("source_plane_distance",    self.source_plane_distance)
                print("image_plane_distance",     self.image_plane_distance)
                print("angles_respect_to",        self.angles_respect_to)
                print("incidence_angle_deg",      self.incidence_angle_deg)
                print("reflection_angle_deg",     self.reflection_angle_deg)
                print("mirror_orientation_angle", self.mirror_orientation_angle)


if __name__ == "__main__":
    a = QApplication(sys.argv)
    ow = ConicCoefficientsMirror()
    ow.show()
    a.exec_()
    ow.saveSettings()



import os, sys
import numpy
import scipy.constants as codata


from syned.storage_ring.magnetic_structures.undulator import Undulator
from syned.storage_ring.magnetic_structures.wiggler import Wiggler

from syned.storage_ring.magnetic_structures import insertion_device

from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QMessageBox, QApplication
from PyQt5.QtCore import QRect

from orangewidget import gui
from orangewidget import widget
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.widgets.gui import ConfirmDialog

from syned.storage_ring.light_source import LightSource, ElectronBeam
from syned.beamline.beamline import Beamline

from oasys.widgets.gui import ConfirmDialog

import orangecanvas.resources as resources


m2ev = codata.c * codata.h / codata.e

VERTICAL = 1
HORIZONTAL = 2
BOTH = 3


class OWEBS(OWWidget):

    name = "ESRF-EBS ID Light Source"
    description = "Syned: ESRF-EBS ID Light Source"
    icon = "icons/ebs.png"
    priority = 1


    maintainer = "Manuel Sanchez del Rio"
    maintainer_email = "srio(@at@)esrf.eu"
    category = "ESRF-EBS Syned Light Sources"
    keywords = ["data", "file", "load", "read"]

    outputs = [{"name":"SynedData",
                "type":Beamline,
                "doc":"Syned Beamline",
                "id":"data"}]


    want_main_area = 1


    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 645

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 650

    TABS_AREA_HEIGHT = 625
    CONTROL_AREA_WIDTH = 450


    electron_energy_in_GeV = Setting(6.0)
    electron_energy_spread = Setting(0.001)
    ring_current           = Setting(0.2)
    number_of_bunches      = Setting(0.0)

    moment_xx           = Setting(0.0)
    moment_xxp          = Setting(0.0)
    moment_xpxp         = Setting(0.0)
    moment_yy           = Setting(0.0)
    moment_yyp          = Setting(0.0)
    moment_ypyp         = Setting(0.0)

    electron_beam_size_h       = Setting(0.0)
    electron_beam_divergence_h = Setting(0.0)
    electron_beam_size_v       = Setting(0.0)
    electron_beam_divergence_v = Setting(0.0)

    electron_beam_emittance_h = Setting(0.0)
    electron_beam_emittance_v = Setting(0.0)
    electron_beam_beta_h = Setting(0.0)
    electron_beam_beta_v = Setting(0.0)
    electron_beam_alpha_h = Setting(0.0)
    electron_beam_alpha_v = Setting(0.0)
    electron_beam_eta_h = Setting(0.0)
    electron_beam_eta_v = Setting(0.0)
    electron_beam_etap_h = Setting(0.0)
    electron_beam_etap_v = Setting(0.0)

    type_of_properties = Setting(1)

    auto_energy = Setting(0.0)
    auto_harmonic_number = Setting(1)

    K_horizontal       = Setting(0.5)
    K_vertical         = Setting(0.5)
    period_length      = Setting(0.018)
    number_of_periods  = Setting(10)

    ebs_id_index = Setting(0)
    gap_mm = Setting(0.0)

    gap_min = Setting(5.0)
    gap_max = Setting(30.0)
    harmonic_max = Setting(3)

    a0 = Setting('2.083')
    a1 = Setting('1.0054')
    a2 = Setting('')
    a3 = Setting('')
    a4 = Setting('')
    a5 = Setting('')
    a6 = Setting('')

    # data_url = 'ftp://ftp.esrf.eu/pub/scisoft/syned/resources/jsrund.csv'
    # create it in nice with the ID app: /segfs/tango/bin/jsrund
    data_url = os.path.join(resources.package_dirname("orangecontrib.esrf.syned.data"), 'jsrund.csv')

    data_dict = None

    def __init__(self):

        self.get_data_dictionary_csv()

        # OLD FORMAT
        # self.data_url = "https://raw.githubusercontent.com/srio/shadow3-scripts/master/ESRF-LIGHTSOURCES-EBS/ebs_ids.json"
        # self.get_data_dictionary()

        self.runaction = widget.OWAction("Send Data", self)
        self.runaction.triggered.connect(self.send_data)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Send Data", callback=self.send_data)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        gui.separator(self.controlArea)

        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width()*0.05),
                               round(geom.height()*0.05),
                               round(min(geom.width()*0.98, self.MAX_WIDTH)),
                               round(min(geom.height()*0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        self.tabs_setting = oasysgui.tabWidget(self.controlArea)
        self.tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        self.tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        self.tab_sou = oasysgui.createTabPage(self.tabs_setting, "Light Source Setting")

        gui.comboBox(self.tab_sou, self, "ebs_id_index", label="Load ID parameters from database list: ", labelWidth=350,
                     items=self.get_id_list(), callback=self.set_id, sendSelectedValue=False, orientation="horizontal")

        self.electron_beam_box = oasysgui.widgetBox(self.tab_sou, "Electron Beam/Machine Parameters", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.electron_beam_box, self, "electron_energy_in_GeV", "Energy [GeV]",  labelWidth=260, valueType=float, orientation="horizontal", callback=self.update)
        oasysgui.lineEdit(self.electron_beam_box, self, "electron_energy_spread", "Energy Spread", labelWidth=260, valueType=float, orientation="horizontal", callback=self.update)
        oasysgui.lineEdit(self.electron_beam_box, self, "ring_current", "Ring Current [A]",        labelWidth=260, valueType=float, orientation="horizontal", callback=self.update)

        gui.comboBox(self.electron_beam_box, self, "type_of_properties", label="Electron Beam Properties", labelWidth=350,
                     items=["From 2nd Moments", "From Size/Divergence", "From Twiss papameters","Zero emittance", "EBS (S28D)"],
                     callback=self.update_electron_beam,
                     sendSelectedValue=False, orientation="horizontal")

        self.left_box_2_1 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xx",   "<x x>   [m^2]",   labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xxp",  "<x x'>  [m.rad]", labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_xpxp", "<x' x'> [rad^2]", labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yy",   "<y y>   [m^2]",   labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_yyp",  "<y y'>  [m.rad]", labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_1, self, "moment_ypyp", "<y' y'> [rad^2]", labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)


        self.left_box_2_2 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="vertical", height=150)

        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_h",       "Horizontal Beam Size \u03c3x [m]",          labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_size_v",       "Vertical Beam Size \u03c3y [m]",            labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_h", "Horizontal Beam Divergence \u03c3'x [rad]", labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_2, self, "electron_beam_divergence_v", "Vertical Beam Divergence \u03c3'y [rad]",   labelWidth=260, valueType=float, orientation="horizontal",  callback=self.update)

        self.left_box_2_3 = oasysgui.widgetBox(self.electron_beam_box, "", addSpace=False, orientation="horizontal",height=150)
        self.left_box_2_3_l = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        self.left_box_2_3_r = oasysgui.widgetBox(self.left_box_2_3, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_emittance_h", "\u03B5x [m.rad]",labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_alpha_h",     "\u03B1x",        labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_beta_h",      "\u03B2x [m]",    labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_eta_h",       "\u03B7x",        labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_l, self, "electron_beam_etap_h",      "\u03B7'x",       labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)


        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_emittance_v", "\u03B5y [m.rad]",labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_alpha_v",     "\u03B1y",        labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_beta_v",      "\u03B2y [m]",    labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_eta_v",       "\u03B7y",        labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)
        oasysgui.lineEdit(self.left_box_2_3_r, self, "electron_beam_etap_v",      "\u03B7'y",       labelWidth=75, valueType=float, orientation="horizontal",  callback=self.update)

        gui.rubber(self.controlArea)

        ###################

        left_box_1 = oasysgui.widgetBox(self.tab_sou, "ID Parameters", addSpace=True, orientation="vertical")

        oasysgui.lineEdit(left_box_1, self, "period_length", "Period Length [m]", labelWidth=260,
                          valueType=float, orientation="horizontal", callback=self.update)
        oasysgui.lineEdit(left_box_1, self, "number_of_periods", "Number of Periods", labelWidth=260,
                          valueType=float, orientation="horizontal", callback=self.update)



        left_box_1 = oasysgui.widgetBox(self.tab_sou, "Setting", addSpace=True, orientation="vertical")

        # oasysgui.lineEdit(left_box_1, self, "K_horizontal", "Horizontal K", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(left_box_1, self, "K_vertical", "Vertical K", labelWidth=260,
                          valueType=float, orientation="horizontal", callback=self.set_K)

        oasysgui.lineEdit(left_box_1, self, "gap_mm", "Undulator Gap [mm]",
                          labelWidth=250, valueType=float, orientation="horizontal",
                          callback=self.set_gap)

        left_box_2 = oasysgui.widgetBox(left_box_1, "", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_2, self, "auto_energy", "Photon Energy [eV]",
                          labelWidth=250, valueType=float, orientation="horizontal",
                          callback=self.auto_set_undulator_V)
        oasysgui.lineEdit(left_box_2, self, "auto_harmonic_number", "Harmonic",
                          labelWidth=250, valueType=int, orientation="horizontal",
                          callback=self.auto_set_undulator_V)

        ####################################################


        tab_util = oasysgui.createTabPage(self.tabs_setting, "Settings")

        left_box_0 = oasysgui.widgetBox(tab_util, "Advanced settings",
                        addSpace=False, orientation="vertical", height=450)

        oasysgui.lineEdit(left_box_0, self, "gap_min",  "minimum gap",
                          labelWidth=260, valueType=float, orientation="horizontal",
                          callback=self.update)

        oasysgui.lineEdit(left_box_0, self, "gap_max",  "maximum gap (for plots)",
                          labelWidth=260, valueType=float, orientation="horizontal",
                          callback=self.update)

        oasysgui.lineEdit(left_box_0, self, "harmonic_max",  "maximum harmonic (for plots)",
                          labelWidth=260, valueType=int, orientation="horizontal",
                          callback=self.update)

        left_box_00 = oasysgui.widgetBox(left_box_0, "Gap parametrization", addSpace=False, orientation="vertical")
        oasysgui.lineEdit(left_box_00, self, "a0", "a0", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a1", "a1", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a2", "a2", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a3", "a3", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a4", "a4", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a5", "a5", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)
        oasysgui.lineEdit(left_box_00, self, "a6", "a6", labelWidth=260, valueType=str, orientation="horizontal", callback=self.set_K)

        self.initializeTabs()

        # self.populate_gap_parametrization()
        # self.populate_electron_beam()
        # self.populate_magnetic_structure()
        # self.set_ebs_electron_beam()

        self.populate_settings_after_setting_K()
        self.set_visible()
        self.update()


    def get_id_list(self):
        out_list = [("ID%02d %s" % (self.data_dict["straight_section"][i], self.data_dict["id_name"][i])) for i in
                    range(len(self.data_dict["id_name"]))]

        out_list.insert(0,"<None>") # We add None at the beginning: ebs_id_index is the dict index plus one
        return out_list

    def titles(self):
        return ["K vs Gap", "B vs Gap", "Gap vs resonance energy", "Power vs Gap", "Power density peak at 30 m vs Gap"]

    def xtitles(self):
        return ['Gap [mm]'] * len(self.titles())

    def ytitles(self):
        return ['K', 'B [T]', 'Photon energy [eV]', 'Power [W]', 'Power density peak @ 30 m [W/mm$^2$]']

    def initializeTabs(self):
        self.tabs = oasysgui.tabWidget(self.mainArea)

        self.tab = [oasysgui.createTabPage(self.tabs, "Info",),
                    oasysgui.createTabPage(self.tabs, "K vs Gap"),
                    oasysgui.createTabPage(self.tabs, "B vs Gap"),
                    oasysgui.createTabPage(self.tabs, "Resonance vs Gap"),
                    oasysgui.createTabPage(self.tabs, "Power vs Gap"),
                    oasysgui.createTabPage(self.tabs, "Pow_dens_peak vs Gap"),
                    ]

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)


        self.info_id = oasysgui.textArea(height=self.IMAGE_HEIGHT-5, width=self.IMAGE_WIDTH-5)
        profile_box = oasysgui.widgetBox(self.tab[0], "", addSpace=True, orientation="horizontal",
                                         height = self.IMAGE_HEIGHT, width=self.IMAGE_WIDTH-5)
        profile_box.layout().addWidget(self.info_id)


        n_plots = len(self.titles())
        self.plot_canvas = [None] * (1 + n_plots)


        for i in range(n_plots):
            self.plot_canvas[i] = oasysgui.plotWindow(roi=False, control=False, position=True)
            self.plot_canvas[i].setDefaultPlotLines(True)
            self.plot_canvas[i].setActiveCurveColor(color='blue')
            self.plot_canvas[i].setGraphXLabel(self.xtitles()[i])
            self.plot_canvas[i].setGraphYLabel(self.ytitles()[i])
            self.plot_canvas[i].setGraphTitle(self.titles()[i])
            self.plot_canvas[i].setInteractiveMode(mode='zoom')


        for index in range(0, 5):
            self.tab[index + 1].layout().addWidget(self.plot_canvas[index])

        self.tabs.setCurrentIndex(1)

    def check_magnetic_structure(self):
        congruence.checkPositiveNumber(self.K_horizontal, "Horizontal K")
        congruence.checkPositiveNumber(self.K_vertical, "Vertical K")
        congruence.checkStrictlyPositiveNumber(self.period_length, "Period Length")
        congruence.checkStrictlyPositiveNumber(self.number_of_periods, "Number of Periods")

    def set_ebs_electron_beam(self):
        self.type_of_properties = 1
        self.electron_beam_size_h = 30.1836e-6
        self.electron_beam_size_v = 3.63641e-6
        self.electron_beam_divergence_h = 4.36821e-6
        self.electron_beam_divergence_v = 1.37498e-6

        #
        eb = self.get_electron_beam()

        moment_xx, moment_xxp, moment_xpxp, moment_yy, moment_yyp, moment_ypyp = eb.get_moments_all()
        self.moment_xx   = moment_xx
        self.moment_yy   = moment_yy
        self.moment_xxp  = moment_xxp
        self.moment_yyp  = moment_yyp
        self.moment_xpxp = moment_xpxp
        self.moment_ypyp = moment_ypyp

        ex, ax, bx, ey, ay, by = eb.get_twiss_no_dispersion_all()
        self.electron_beam_beta_h = bx
        self.electron_beam_beta_v = by
        self.electron_beam_alpha_h = ax
        self.electron_beam_alpha_v = ay
        self.electron_beam_eta_h = ex
        self.electron_beam_eta_v = ey
        self.electron_beam_etap_h = 0.0
        self.electron_beam_etap_v = 0.0
        self.electron_beam_emittance_h = 1.3166e-10
        self.electron_beam_emittance_v = 5e-12

    def update_electron_beam(self):
        if self.type_of_properties == 4:
            self.set_ebs_electron_beam()
        self.set_visible()
        self.update()

    def update(self):
        self.check_data()
        self.update_info()
        self.update_plots()

    def update_info(self):

        syned_electron_beam = self.get_electron_beam()
        syned_undulator = self.get_magnetic_structure()

        gamma = self.gamma()

        if self.ebs_id_index == 0:
            id = "<None>"
        else:
            id = "ID%02d %s" % (self.data_dict["straight_section"][self.ebs_id_index-1], self.data_dict["id_name"][self.ebs_id_index-1])

        info_parameters = {
            "electron_energy_in_GeV":self.electron_energy_in_GeV,
            "gamma":"%8.3f"%self.gamma(),
            "ring_current":"%4.3f "%syned_electron_beam.current(),
            "K_horizontal":syned_undulator.K_horizontal(),
            "K_vertical": syned_undulator.K_vertical(),
            "period_length": syned_undulator.period_length(),
            "number_of_periods": syned_undulator.number_of_periods(),
            "undulator_length": syned_undulator.length(),
            "resonance_energy":"%6.3f"%syned_undulator.resonance_energy(gamma,harmonic=1),
            "resonance_energy3": "%6.3f" % syned_undulator.resonance_energy(gamma,harmonic=3),
            "resonance_energy5": "%6.3f" % syned_undulator.resonance_energy(gamma,harmonic=5),
            "B_horizontal":"%4.2F"%syned_undulator.magnetic_field_horizontal(),
            "B_vertical": "%4.2F" % syned_undulator.magnetic_field_vertical(),
            "cc_1": "%4.2f" % (1e6*syned_undulator.gaussian_central_cone_aperture(gamma,1)),
            "cc_3": "%4.2f" % (1e6*syned_undulator.gaussian_central_cone_aperture(gamma,3)),
            "cc_5": "%4.2f" % (1e6*syned_undulator.gaussian_central_cone_aperture(gamma,5)),
            # "cc_7": "%4.2f" % (self.gaussian_central_cone_aperture(7)*1e6),
            "sigma_rad": "%5.2f"        % (1e6*syned_undulator.get_sigmas_radiation(gamma,harmonic=1)[0]),
            "sigma_rad_prime": "%5.2f"  % (1e6*syned_undulator.get_sigmas_radiation(gamma,harmonic=1)[1]),
            "sigma_rad3": "%5.2f"       % (1e6*syned_undulator.get_sigmas_radiation(gamma,harmonic=3)[0]),
            "sigma_rad_prime3": "%5.2f" % (1e6*syned_undulator.get_sigmas_radiation(gamma,harmonic=3)[1]),
            "sigma_rad5": "%5.2f" % (1e6 * syned_undulator.get_sigmas_radiation(gamma, harmonic=5)[0]),
            "sigma_rad_prime5": "%5.2f" % (1e6 * syned_undulator.get_sigmas_radiation(gamma, harmonic=5)[1]),
            "first_ring_1": "%5.2f" % (1e6*syned_undulator.get_resonance_ring(gamma, harmonic=1, ring_order=1)),
            "first_ring_3": "%5.2f" % (1e6*syned_undulator.get_resonance_ring(gamma, harmonic=3, ring_order=1)),
            "first_ring_5": "%5.2f" % (1e6*syned_undulator.get_resonance_ring(gamma, harmonic=5, ring_order=1)),
            "Sx": "%5.2f"  % (1e6*syned_undulator.get_photon_sizes_and_divergences(syned_electron_beam)[0]),
            "Sy": "%5.2f"  % (1e6*syned_undulator.get_photon_sizes_and_divergences(syned_electron_beam)[1]),
            "Sxp": "%5.2f" % (1e6*syned_undulator.get_photon_sizes_and_divergences(syned_electron_beam)[2]),
            "Syp": "%5.2f" % (1e6*syned_undulator.get_photon_sizes_and_divergences(syned_electron_beam)[3]),
            "und_power": "%5.2f" % syned_undulator.undulator_full_emitted_power(gamma,syned_electron_beam.current()),
            "CF_h": "%4.3f" % syned_undulator.approximated_coherent_fraction_horizontal(syned_electron_beam,harmonic=1),
            "CF_v": "%4.3f" % syned_undulator.approximated_coherent_fraction_vertical(syned_electron_beam,harmonic=1),
            "CF": "%4.3f" % syned_undulator.approximated_coherent_fraction(syned_electron_beam,harmonic=1),
            "url": self.data_url,
            "id": id,
            "gap": "%4.3f" % self.calculate_gap_from_K(),
            "a0": "%s" % str(self.a0),
            "a1": "%s" % str(self.a1),
            "a2": "%s" % str(self.a2),
            "a3": "%s" % str(self.a3),
            "a4": "%s" % str(self.a4),
            "a5": "%s" % str(self.a5),
            "a6": "%s" % str(self.a6),
            }

        self.info_id.setText(self.info_template().format_map(info_parameters))
        # self.tabs[0].setText(self.info_template().format_map(info_parameters))


    def info_template(self):
        return \
"""
data url: {url}
id_name: {id}

================ input parameters ===========
Electron beam energy [GeV]: {electron_energy_in_GeV}
Electron current:           {ring_current}
Period Length [m]:          {period_length}
Number of Periods:          {number_of_periods}

Horizontal K:               {K_horizontal}
Vertical K:                 {K_vertical}
==============================================

Electron beam gamma:                {gamma}
Undulator Length [m]:               {undulator_length}
Horizontal Peak Magnetic field [T]: {B_horizontal}
Vertical Peak Magnetic field [T]:   {B_vertical}

Total power radiated by the undulator [W]: {und_power}

Gap in use: {gap} mm
Using gap parametrization: 
    a0: {a0}
    a1: {a1}
    a2: {a2}
    a3: {a3}
    a4: {a4}
    a5: {a5}
    a6: {a6}

Note on calculation: 
A = [a0,a1,a2,...] = [B_1, B_2,..., alpha_1, alpha_2,...]
Bmax = Sum[B_i * exp( -pi * alpha_i * (gap[mm] / id_period[mm]) )] with i from 1 to the semilength of A. 

Resonances:

Photon energy [eV]: 
       for harmonic 1 : {resonance_energy}
       for harmonic 3 : {resonance_energy3}
       for harmonic 5 : {resonance_energy5}

Central cone (RMS urad):
       for harmonic 1 : {cc_1}
       for harmonic 3 : {cc_3}
       for harmonic 5 : {cc_5}

First ring at (urad):
       for harmonic 1 : {first_ring_1}
       for harmonic 3 : {first_ring_3}
       for harmonic 5 : {first_ring_5}

Sizes and divergences of radiation :
    at 1st harmonic: sigma: {sigma_rad} um, sigma': {sigma_rad_prime} urad
    at 3rd harmonic: sigma: {sigma_rad3} um, sigma': {sigma_rad_prime3} urad
    at 5th harmonic: sigma: {sigma_rad5} um, sigma': {sigma_rad_prime5} urad
    
Sizes and divergences of photon source (convolution) at resonance (1st harmonic): :
    Sx: {Sx} um
    Sy: {Sy} um,
    Sx': {Sxp} urad
    Sy': {Syp} urad
    
Approximated coherent fraction at 1st harmonic: 
    Horizontal: {CF_h}
    Vertical: {CF_v} 
    Coherent fraction 2D (HxV): {CF} 

"""

    def get_magnetic_structure(self, check_for_wiggler=False):

        if not(check_for_wiggler):
            return Undulator(K_horizontal=self.K_horizontal,
                             K_vertical=self.K_vertical,
                             period_length=self.period_length,
                             number_of_periods=self.number_of_periods)
        else:
            id_name = self.get_id_list()[self.ebs_id_index]
            if "W" in id_name:
                return Wiggler(K_horizontal=self.K_horizontal,
                                 K_vertical=self.K_vertical,
                                 period_length=self.period_length,
                                 number_of_periods=self.number_of_periods)
            else:
                return Undulator(K_horizontal=self.K_horizontal,
                                 K_vertical=self.K_vertical,
                                 period_length=self.period_length,
                                 number_of_periods=self.number_of_periods)


    def check_magnetic_structure_instance(self, magnetic_structure):
        if not isinstance(magnetic_structure, Undulator):
            raise ValueError("Magnetic Structure is not a Undulator")

    def populate_magnetic_structure(self):
        # if magnetic_structure is None:
        index = self.ebs_id_index - 1
        self.K_horizontal = 0.0
        self.K_vertical = numpy.round(self.data_dict["Kmax"][index],4)
        self.period_length = numpy.round(self.data_dict["id_period"][index],4)
        self.number_of_periods = numpy.round(self.data_dict["id_length"][index] / self.period_length,3)

    def populate_gap_parametrization(self):
        index = self.ebs_id_index - 1
        if self.data_dict["a0"][index] is None:
            self.a0 = ''
        else:
            self.a0 = self.data_dict["a0"][index]


        if self.data_dict["a1"][index] is None:
            self.a1 = ''
        else:
            self.a1 = self.data_dict["a1"][index]

        if self.data_dict["a2"][index] is None:
            self.a2 = ''
        else:
            self.a2 = self.data_dict["a2"][index]

        if self.data_dict["a3"][index] is None:
            self.a3 = ''
        else:
            self.a3 = self.data_dict["a3"][index]

        if self.data_dict["a4"][index] is None:
            self.a4 = ''
        else:
            self.a4 = self.data_dict["a4"][index]

        if self.data_dict["a5"][index] is None:
            self.a5 = ''
        else:
            self.a5= self.data_dict["a5"][index]

        if self.data_dict["a6"][index] is None:
            self.a6 = ''
        else:
            self.a6 = self.data_dict["a6"][index]


        # self.a0 = self.data_dict["a0"][index]
        # self.a1 = self.data_dict["a1"][index]
        # self.a2 = self.data_dict["a2"][index]
        # self.a3 = self.data_dict["a3"][index]
        # self.a4 = self.data_dict["a4"][index]
        # self.a5 = self.data_dict["a5"][index]
        # self.a6 = self.data_dict["a6"][index]

    def populate_settings_after_setting_K(self):

        syned_undulator = self.get_magnetic_structure()
        self.auto_energy = numpy.round(syned_undulator.resonance_energy(self.gamma(),
                                            harmonic=self.auto_harmonic_number),3)

        self.gap_mm = numpy.round(self.calculate_gap_from_K(), 3)

    def set_gap(self, which=VERTICAL):
        if self.gap_mm < self.gap_min:
            if ConfirmDialog.confirmed(self, message="Gap is smaller than minimum. Set to minimum?"):
                self.gap_mm = self.gap_min

        if self.gap_mm > self.gap_max:
            if ConfirmDialog.confirmed(self, message="Gap is larger than maximum. Set to maximum?"):
                self.gap_mm = self.gap_max

        if self.gap_mm < self.gap_min:
            raise Exception("Gap is smaller than minimum")

        if self.gap_mm > self.gap_max:
            raise Exception("Gap is larger than maximum")

        K = numpy.round(self.calculate_K_from_gap(), 3)

        if which == VERTICAL:
            self.K_vertical = K
            self.K_horizontal = 0.0

        if which == BOTH:
            Kboth = round(K / numpy.sqrt(2), 6)
            self.K_vertical =  Kboth
            self.K_horizontal = Kboth

        if which == HORIZONTAL:
            self.K_horizontal = K
            self.K_vertical = 0.0

        self.populate_settings_after_setting_K()
        self.update()

    def set_id(self):
        if self.ebs_id_index !=0:
            self.populate_gap_parametrization()
            self.populate_magnetic_structure()
            self.gap_min = self.data_dict["id_minimum_gap_mm"][self.ebs_id_index-1]

        self.populate_settings_after_setting_K()
        self.update()

    def set_K(self):
        self.populate_settings_after_setting_K()
        self.update()

    def auto_set_undulator_V(self):
        self.set_resonance_energy(VERTICAL)

    def auto_set_undulator_H(self):
        self.set_resonance_energy(HORIZONTAL)

    def auto_set_undulator_B(self):
        self.set_resonance_energy(BOTH)

    def set_resonance_energy(self, which=VERTICAL):
        congruence.checkStrictlyPositiveNumber(self.auto_energy, "Set Undulator at Energy")
        congruence.checkStrictlyPositiveNumber(self.auto_harmonic_number, "As Harmonic #")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV, "Energy")
        congruence.checkStrictlyPositiveNumber(self.period_length, "Period Length")

        wavelength = self.auto_harmonic_number*m2ev/self.auto_energy
        K = round(numpy.sqrt(2*(((wavelength*2*self.gamma()**2)/self.period_length)-1)), 6)

        Kmax = self.calculate_K_from_gap(self.gap_min)
        Kmin = self.calculate_K_from_gap(self.gap_max)

        if numpy.isnan(K):
            if ConfirmDialog.confirmed(self, message="Impossible configuration. Set to Kmin=%f?" % (Kmin)):
                K = numpy.round(Kmin,4)

        if (K > Kmax):
            if ConfirmDialog.confirmed(self, message="Needed K (%f) > Kmax (%f). Reset to Kmax?" % (K, Kmax)):
                K = numpy.round(Kmax,4)

        if (K < Kmin):
            if ConfirmDialog.confirmed(self, message="Needed K (%f) < Kmin (%f). Reset to Kmin?" % (K, Kmin)):
                K = numpy.round(Kmin,4)

        if which == VERTICAL:
            self.K_vertical = K
            self.K_horizontal = 0.0

        if which == BOTH:
            Kboth = round(K / numpy.sqrt(2), 6)
            self.K_vertical =  Kboth
            self.K_horizontal = Kboth

        if which == HORIZONTAL:
            self.K_horizontal = K
            self.K_vertical = 0.0

        self.populate_settings_after_setting_K()
        self.update()

    def plot_graph(self, plot_canvas_index, curve_name, x_values, y_values, xtitle="", ytitle="",
                   color='blue', replace=True):
        self.plot_canvas[plot_canvas_index].addCurve(x_values, y_values, curve_name, symbol='', color=color, replace=replace) #'+', '^', ','
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
        self.plot_canvas[plot_canvas_index].replot()

    def update_plots(self):

        gap_mm = numpy.linspace(self.gap_min * 0.9, self.gap_max * 1.1, 1000)

        Karray = self.calculate_K_from_gap(gap_mm)
        Karray_horizontal = numpy.zeros_like(Karray)

        Bfield = Karray / (self.period_length * codata.e / (2 * numpy.pi * codata.m_e * codata.c))

        E1_array = self.calculate_resonance_energy(Karray)

        ptot = (self.number_of_periods /6) * codata.value('characteristic impedance of vacuum') * \
               self.ring_current * codata.e * 2 * numpy.pi * codata.c * self.gamma()**2 * \
               (Karray**2 + Karray_horizontal**2) / self.period_length

        ### power density peak at 30 m, not yet implemented for helical undulators ###
        dist_to_source = 30
        
        ### From: Undulators, Wigglers and their Applications - H. Onuki & P Elleaume ###
        ### Chapter 3: Undulator radiation eqs. 68 and 69 ###
        g_k = Karray * ((Karray**6) + (24/7)*(Karray**4) + 4*(Karray**2) + (16/7))/((1 + (Karray**2))**(7/2))

        p_dens_peak = ((21 * (self.gamma()**2)) / (16 * numpy.pi * Karray) * ptot * g_k)/((dist_to_source * 1e3)**2)

        self.plot_graph(0, self.titles()[0], gap_mm, Karray, xtitle=self.xtitles()[0], ytitle=self.ytitles()[0])
        self.plot_graph(1, self.titles()[1], gap_mm, Bfield, xtitle=self.xtitles()[1], ytitle=self.ytitles()[1])
        #
        #
        #
        xtitle = "Photon energy [keV]"
        ytitle = "Gap [mm]"
        colors = ['green', 'black', 'red', 'brown', 'orange', 'pink'] * self.harmonic_max
        for i in range(1, self.harmonic_max+1):
            self.plot_canvas[2].addCurve(E1_array * i* 1e-3, gap_mm,
                                         "harmonic %d" % i,
                                         xlabel=xtitle, ylabel=ytitle,
                                         symbol='', color=colors[i-1])


        self.plot_canvas[2].getLegendsDockWidget().setFixedHeight(150)
        self.plot_canvas[2].getLegendsDockWidget().setVisible(True)
        self.plot_canvas[2].setActiveCurve("harmonic 1")
        self.plot_canvas[2].replot()
        #
        #
        #
        self.plot_graph(3, self.titles()[3], gap_mm, ptot, xtitle=self.xtitles()[3], ytitle=self.ytitles()[3])

        self.plot_graph(4, self.titles()[4], gap_mm, p_dens_peak, xtitle=self.xtitles()[4], ytitle=self.ytitles()[4])


    def calculate_resonance_energy(self, Karray):

        theta_x = 0.0
        theta_z = 0.0
        K_vertical = Karray
        K_horizontal = numpy.zeros_like(K_vertical)

        wavelength = (self.period_length / (2.0 * self.gamma() ** 2)) * \
                         (1 + K_vertical ** 2 / 2.0 + K_horizontal ** 2 / 2.0 + \
                          self.gamma() ** 2 * (theta_x ** 2 + theta_z ** 2))

        wavelength /= self.auto_harmonic_number

        frequency = codata.c / wavelength * self.auto_harmonic_number

        energy_in_ev = codata.h * frequency / codata.e

        return energy_in_ev * self.auto_harmonic_number


    def calculate_K_from_gap(self, gap_mm=None):

        if gap_mm is None: gap_mm = self.gap_mm
        id_period_mm  = self.period_length * 1e3
        id_name = self.get_id_list()[self.ebs_id_index]

        try:
            a0 = float(self.a0)
        except:
            a0 = None

        try:
            a1 = float(self.a1)
        except:
            a1 = None

        try:
            a2 = float(self.a2)
        except:
            a2 = None

        try:
            a3 = float(self.a3)
        except:
            a3 = None

        try:
            a4 = float(self.a4)
        except:
            a4 = None

        try:
            a5 = float(self.a5)
        except:
            a5 = None

        try:
            a6 = float(self.a6)
        except:
            a6 = None


        Bmax = self.calculate_B_from_gap_and_A_vector(
            gap_mm, id_period_mm, id_name,
            a0=a0, a1=a1, a2=a2, a3=a3, a4=a4, a5=a5, a6=a6)

        Kmax = Bmax * (id_period_mm * 1e-3) * codata.e / (2 * numpy.pi * codata.m_e * codata.c)

        return Kmax

    def calculate_gap_from_K(self, Kvalue=None):
        if Kvalue is None: Kvalue = self.K_vertical

        gap_mm_array = numpy.linspace( self.gap_min * 0.9, self.gap_max * 1.1, 1000)
        K_array = self.calculate_K_from_gap(gap_mm_array)

        # if ((Kvalue < K_array.min()) or (Kvalue > K_array.max())):
        #     print("K: %g, gap_max: %g; K_array min:%g, max:%g" % (Kvalue, self.gap_max, K_array.min(), K_array.max()))
        #     raise Exception("Cannot interpolate...")

        if Kvalue < K_array.min():
            if ConfirmDialog.confirmed(self, message="K=%g is smaller than minimum=%g. Set to minimum?" % (Kvalue, K_array.min())):
                Kvalue = K_array.min()
                self.K_vertical = Kvalue
                self.populate_settings_after_setting_K()
                self.update()

        if Kvalue > K_array.max():
            if ConfirmDialog.confirmed(self, message="K=%g is larger than maximum=%g. Set to maximum?" % (Kvalue, K_array.max())):
                Kvalue = K_array.max()
                self.K_vertical = Kvalue
                self.populate_settings_after_setting_K()
                self.update()

        gap_interpolated = numpy.interp(Kvalue, numpy.flip(K_array), numpy.flip(gap_mm_array))

        return gap_interpolated

    def gamma(self):
        return 1e9*self.electron_energy_in_GeV / (codata.m_e *  codata.c**2 / codata.e)

    def set_visible(self):
        self.left_box_2_1.setVisible(self.type_of_properties == 0)
        self.left_box_2_2.setVisible(self.type_of_properties == 1)
        self.left_box_2_3.setVisible(self.type_of_properties == 2)

    def check_data(self):
        congruence.checkStrictlyPositiveNumber(self.electron_energy_in_GeV , "Energy")
        congruence.checkStrictlyPositiveNumber(self.electron_energy_spread, "Energy Spread")
        congruence.checkStrictlyPositiveNumber(self.ring_current, "Ring Current")

        if self.type_of_properties == 0:
            congruence.checkPositiveNumber(self.moment_xx   , "Moment xx")
            congruence.checkPositiveNumber(self.moment_xpxp , "Moment xpxp")
            congruence.checkPositiveNumber(self.moment_yy   , "Moment yy")
            congruence.checkPositiveNumber(self.moment_ypyp , "Moment ypyp")
        elif self.type_of_properties == 1:
            congruence.checkPositiveNumber(self.electron_beam_size_h       , "Horizontal Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_divergence_h , "Vertical Beam Size")
            congruence.checkPositiveNumber(self.electron_beam_size_v       , "Horizontal Beam Divergence")
            congruence.checkPositiveNumber(self.electron_beam_divergence_v , "Vertical Beam Divergence")
        elif self.type_of_properties == 2:
            congruence.checkPositiveNumber(self.electron_beam_emittance_h, "Horizontal Beam Emittance")
            congruence.checkPositiveNumber(self.electron_beam_emittance_v, "Vertical Beam Emittance")
            congruence.checkNumber(self.electron_beam_alpha_h, "Horizontal Beam Alpha")
            congruence.checkNumber(self.electron_beam_alpha_v, "Vertical Beam Alpha")
            congruence.checkNumber(self.electron_beam_beta_h, "Horizontal Beam Beta")
            congruence.checkNumber(self.electron_beam_beta_v, "Vertical Beam Beta")
            congruence.checkNumber(self.electron_beam_eta_h, "Horizontal Beam Dispersion Eta")
            congruence.checkNumber(self.electron_beam_eta_v, "Vertical Beam Dispersion Eta")
            congruence.checkNumber(self.electron_beam_etap_h, "Horizontal Beam Dispersion Eta'")
            congruence.checkNumber(self.electron_beam_etap_v, "Vertical Beam Dispersion Eta'")

        self.check_magnetic_structure()

    def send_data(self):
        try:
            self.check_data()
            self.send("SynedData", Beamline(light_source=self.get_light_source(check_for_wiggler=True)))
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e.args[0]), QMessageBox.Ok)

            self.setStatusMessage("")
            self.progressBarFinished()

    def get_electron_beam(self):
        electron_beam = ElectronBeam(energy_in_GeV=self.electron_energy_in_GeV,
                                     energy_spread=self.electron_energy_spread,
                                     current=self.ring_current,
                                     number_of_bunches=self.number_of_bunches)

        if self.type_of_properties == 0:
            electron_beam.set_moments_horizontal(self.moment_xx,self.moment_xxp,self.moment_xpxp)
            electron_beam.set_moments_vertical(self.moment_yy, self.moment_yyp, self.moment_ypyp)

        elif self.type_of_properties == 1:
            electron_beam.set_sigmas_all(sigma_x=self.electron_beam_size_h,
                                         sigma_y=self.electron_beam_size_v,
                                         sigma_xp=self.electron_beam_divergence_h,
                                         sigma_yp=self.electron_beam_divergence_v)

        elif self.type_of_properties == 2:
            electron_beam.set_twiss_horizontal(self.electron_beam_emittance_h,
                                             self.electron_beam_alpha_h,
                                             self.electron_beam_beta_h,
                                             self.electron_beam_eta_h,
                                             self.electron_beam_etap_h)
            electron_beam.set_twiss_vertical(self.electron_beam_emittance_v,
                                             self.electron_beam_alpha_v,
                                             self.electron_beam_beta_v,
                                             self.electron_beam_eta_v,
                                             self.electron_beam_etap_v)

        elif self.type_of_properties == 3:
            electron_beam.set_moments_all(0,0,0,0,0,0)

        return electron_beam

    def get_light_source(self, check_for_wiggler=False):
        return LightSource(name=self.get_id_list()[self.ebs_id_index],
                           electron_beam = self.get_electron_beam(),
                           magnetic_structure = self.get_magnetic_structure(check_for_wiggler=check_for_wiggler))

    def callResetSettings(self):
        if ConfirmDialog.confirmed(parent=self, message="Confirm Reset of the Fields?"):
            try:
                self.resetSettings()
            except:
                pass

    def populate_electron_beam(self, electron_beam=None):
        if electron_beam is None:
            electron_beam = ElectronBeam(
                                        energy_in_GeV = 6.0,
                                        energy_spread = 0.001,
                                        current = 0.2,
                                        number_of_bunches = 1,
                                        moment_xx   = (3.01836e-05)**2,
                                        moment_xxp  = (0.0)**2,
                                        moment_xpxp = (4.36821e-06)**2,
                                        moment_yy   = (3.63641e-06)**2,
                                        moment_yyp  = (0.0)**2,
                                        moment_ypyp = (1.37498e-06)**2,
                                        )

        self.electron_energy_in_GeV = electron_beam._energy_in_GeV
        self.electron_energy_spread = electron_beam._energy_spread
        self.ring_current = electron_beam._current
        self.number_of_bunches = electron_beam._number_of_bunches

        self.type_of_properties = 1

        self.moment_xx   = electron_beam._moment_xx
        self.moment_xxp  = electron_beam._moment_xxp
        self.moment_xpxp = electron_beam._moment_xpxp
        self.moment_yy   = electron_beam._moment_yy
        self.moment_yyp  = electron_beam._moment_yyp
        self.moment_ypyp = electron_beam._moment_ypyp

        x, xp, y, yp = electron_beam.get_sigmas_all()

        self.electron_beam_size_h = x
        self.electron_beam_size_v = y
        self.electron_beam_divergence_h = xp
        self.electron_beam_divergence_v = yp

    def calculate_B_from_gap_and_A_vector(self, id_gap_mm, id_period_mm, id_name,
                       a0=None, a1=None, a2=None, a3=None, a4=None, a5=None, a6=None,
                       check_elliptical=True):

        if check_elliptical:
            if "HU" in id_name:
                ConfirmDialog.confirmed(self, message="Helical/Apple undulators not implemented in this app (wrong results)")

        if "CPMU" in id_name:
            B  = a0 * numpy.exp(-numpy.pi * a3 * 1 * id_gap_mm / id_period_mm)
            B += a1 * numpy.exp(-numpy.pi * a4 * 2 * id_gap_mm / id_period_mm)
            B += a2 * numpy.exp(-numpy.pi * a5 * 3 * id_gap_mm / id_period_mm)

        elif "HU" in id_name:  # this is for apple undulator... It is applied also (WRONG!) to helical undulators
            reference_gap = 20.0
            B = a0 * numpy.exp(-numpy.pi * (id_gap_mm - reference_gap) / id_period_mm)
        else:
            if (a2 is None) and (a3 is None): # only one "harmonic"
                B = a0 * numpy.exp(-numpy.pi * a1 * id_gap_mm / id_period_mm)
            else:
                if (a4 is None) and (a5 is None):  # 2 "harmonics"
                    B =  a0 * numpy.exp(-numpy.pi * a2 * 1 * id_gap_mm / id_period_mm)
                    B += a1 * numpy.exp(-numpy.pi * a3 * 2 * id_gap_mm / id_period_mm)
                else: # 3 harmonics
                    B =  a0 * numpy.exp(-numpy.pi * a3 * 1 * id_gap_mm / id_period_mm)
                    B += a1 * numpy.exp(-numpy.pi * a4 * 2 * id_gap_mm / id_period_mm)
                    B += a2 * numpy.exp(-numpy.pi * a5 * 3 * id_gap_mm / id_period_mm)

        return B

    def get_data_dictionary_csv(self):
        url = self.data_url

        # tofloat = lambda s: numpy.array(['0.0' if v == '' else v for v in s]).astype(float)
        tofloat = lambda s: [None if v == '' else float(v) for v in s]
        try:
            filename = url # 'ftp://ftp.esrf.eu/pub/scisoft/syned/resources/jsrund.csv'
            ishift = 1
            a = numpy.genfromtxt(filename, dtype=str, delimiter=',', skip_header=3, skip_footer=1, converters=None, \
                                 missing_values={0: "11.000"}, filling_values={0: "XXX"}, usecols=None, names=None,
                                 excludelist=None, \
                                 deletechars=" !#$%&'()*+, -./:;<=>?@[\]^{|}~", replace_space='', autostrip=True, \
                                 case_sensitive=True, defaultfmt='f%i', unpack=None, usemask=False, loose=True, \
                                 invalid_raise=True, max_rows=None, encoding='bytes')

            number_of_ids = len(a)
            straight_section = a[:, 0].astype(int)
            id_name = a[:, 1]
            id_period = 1e-3 * a[:, 2 + ishift].astype(float)
            id_period_mm = a[:, 2 + ishift].astype(float)
            id_length = 1e-3 * a[:, 3 + ishift].astype(float)

            id_minimum_gap_mm = tofloat(a[:, 4 + ishift])
            for i in range(number_of_ids):
                if id_minimum_gap_mm[i] is None:
                    id_minimum_gap_mm[i] = 30.0 # set to arbitrary value ** Some values are missing!!!**

            a0 = tofloat(a[:, 5 + ishift])
            a1 = tofloat(a[:, 6 + ishift])
            a2 = tofloat(a[:, 7 + ishift])
            a3 = tofloat(a[:, 8 + ishift])
            a4 = tofloat(a[:, 9 + ishift])
            a5 = tofloat(a[:, 10 + ishift])
            a6 = tofloat(a[:, 11 + ishift])


            Bmax = []
            Kmax = []
            for i in range(number_of_ids):
                Bmax_i = self.calculate_B_from_gap_and_A_vector(
                    id_minimum_gap_mm[i], id_period_mm[i], id_name[i],
                    a0=a0[i], a1=a1[i], a2=a2[i], a3=a3[i], a4=a4[i], a5=a5[i], a6=a5[i],
                    check_elliptical=False)
                Bmax.append(Bmax_i)
                Kmax.append(Bmax_i * id_period[i] * codata.e / (2 * numpy.pi * codata.m_e * codata.c))


            out_dict = {}
            out_dict["straight_section"] = straight_section.tolist()
            out_dict["id_name"] = id_name.tolist()
            out_dict["id_minimum_gap_mm"] = id_minimum_gap_mm
            out_dict["Bmax"] = Bmax
            out_dict["Kmax"] = Kmax
            out_dict["straight_section"] = straight_section.tolist()
            out_dict["id_period"] = id_period.tolist()
            out_dict["id_period_mm"] = id_period_mm.tolist()
            out_dict["id_length"] = id_length.tolist()
            out_dict["a0"] = a0
            out_dict["a1"] = a1
            out_dict["a2"] = a2
            out_dict["a3"] = a3
            out_dict["a4"] = a4
            out_dict["a5"] = a5
            out_dict["a6"] = a6

        except:
            out_dict = {}
            out_dict["straight_section"]  = []
            out_dict["id_name"]           = []
            out_dict["id_minimum_gap_mm"] = []
            out_dict["Bmax"]              = []
            out_dict["Kmax"]              = []
            out_dict["straight_section"]  = []
            out_dict["id_period"]         = []
            out_dict["id_period_mm"]      = []
            out_dict["id_length"]         = []
            out_dict["a0"]                = []
            out_dict["a1"]                = []
            out_dict["a2"]                = []
            out_dict["a3"]                = []
            out_dict["a4"]                = []
            out_dict["a5"]                = []
            out_dict["a6"]                = []

        self.data_dict = out_dict

    # OLD data format...
    def get_data_dictionary(self):
        import json
        from urllib.request import urlopen

        file_url = self.data_url # "https://raw.githubusercontent.com/srio/shadow3-scripts/master/ESRF-LIGHTSOURCES-EBS/ebs_ids.json"

        u = urlopen(file_url)
        ur = u.read()
        url = ur.decode(encoding='UTF-8')

        dictionary = json.loads(url)

        self.data_dict = dictionary

if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication

    a = QApplication(sys.argv)
    ow = OWEBS()
    ow.show()
    a.exec_()

    # data_dict = get_data_dictionary()
    # out_list = [("ID%02d %s" % (data_dict["straight_section"][i],data_dict["id_name"][i])) for i in range(len(data_dict["id_name"]))]
    # print(out_list)

    # data_dict_old = data_dict
    # data_dict = get_data_dictionary_csv()
    # out_list = [("ID%02d %s" % (data_dict["straight_section"][i],data_dict["id_name"][i])) for i in range(len(data_dict["id_name"]))]
    # print(out_list)

    # for key in data_dict.keys():
    #     if key != "id_name":
    #         print(numpy.array(data_dict[key]) - numpy.array(data_dict_old[key]))

    # print(data_dict["id_name"])
    # print(data_dict_old["id_name"])



import numpy
from scipy.interpolate import interp2d
import scipy.constants as codata
import xraylib

from oasys.util.oasys_util import write_surface_file, read_surface_file
from oasys.util.oasys_objects import OasysSurfaceData


from syned.beamline.optical_element import OpticalElement
from syned.widget.widget_decorator import WidgetDecorator

from wofry.beamline.decorators import OpticalElementDecorator


# mimics a syned element
class ThinObject(OpticalElement):
    def __init__(self,
                 name="Undefined",
                 file_with_thickness_mesh="",
                 material=""):

        super().__init__(name=name,
                         boundary_shape=None)
        self._material = material
        self._file_with_thickness_mesh = file_with_thickness_mesh

        # support text contaning name of variable, help text and unit. Will be stored in self._support_dictionary
        self._set_support_text([
                    ("name",                       "Name" ,                                "" ),
                    ("boundary_shape",             "Boundary shape",                       "" ),
                    ("material",                   "Material (element, compound or name)", "" ),
                    ("file_with_thickness_mesh",   "File with thickness mesh",             "" ),
            ] )


    def get_material(self):
        return self._material

    def get_file_with_thickness_mesh(self):
        return self._file_with_thickness_mesh

# mimics a wofry element
class WOThinObject(ThinObject, OpticalElementDecorator):
    def __init__(self,
                 name="Undefined",
                 file_with_thickness_mesh="",
                 material="",
                 refraction_index_delta=1e-07,
                 att_coefficient=0.0,
                 ):
        ThinObject.__init__(self,
                      name=name,
                      file_with_thickness_mesh=file_with_thickness_mesh,
                      material=material)

        self._refraction_index_delta = refraction_index_delta
        self._att_coefficient = att_coefficient

    def get_refraction_index(self, photon_energy=10000.0):

        wave_length = codata.h * codata.c / codata.e / photon_energy

        if self.get_material() == "External": # Be
             return self._refraction_index_delta, \
                    self._att_coefficient

        if self.get_material() == "Be": # Be
            element = "Be"
            density = xraylib.ElementDensity(4)
        elif self.get_material() == "Al": # Al
            element = "Al"
            density = xraylib.ElementDensity(13)
        elif self.get_material() == "Diamond": # Diamond
            element = "C"
            density = 3.51
        else:
            raise Exception("Bad material: " + self.get_material())

        refraction_index = xraylib.Refractive_Index(element, photon_energy/1000, density)
        refraction_index_delta = 1 - refraction_index.real
        att_coefficient = 4*numpy.pi * (xraylib.Refractive_Index(element, photon_energy/1000, density)).imag / wave_length

        # print("\n\n\n ==========  parameters recovered from xraylib : ")
        # print("Element: %s" % element)
        # print("        density = %g " % density)
        # print("Photon energy = %g eV" % (photon_energy))
        # print("Refracion index delta = %g " % (refraction_index_delta))
        # print("Attenuation coeff mu = %g m^-1" % (att_coefficient))

        return refraction_index_delta, att_coefficient

    def get_surface_thickness_mesh(self, wavefront):
        xx, yy, zz = read_surface_file(self.get_file_with_thickness_mesh())

        if zz.min() < 0: zz -= zz.min()

        f = interp2d(xx, yy, zz, kind='linear', bounds_error=False, fill_value=0)
        x = wavefront.get_coordinate_x()
        y = wavefront.get_coordinate_y()
        interpolated_profile = f(x, y)
        return x, y, interpolated_profile

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        # return wavefront

        print("\n\n\n ==========  parameters from optical element : ")
        print(self.info())

        photon_energy = wavefront.get_photon_energy()

        refraction_index_delta, att_coefficient = self.get_refraction_index(photon_energy)

        # xx, yy, zz = read_surface_file(self.get_file_with_thickness_mesh())
        # xx, yy, zz = self.get_surface_thickness_mesh()
        # if zz.min() < 0: zz -= zz.min()
        #
        # f = interp2d(xx, yy, zz, kind='linear', bounds_error=False, fill_value=0)
        # x = wavefront.get_coordinate_x()
        # y = wavefront.get_coordinate_y()
        # interpolated_profile = f(x, y)

        x, y, interpolated_profile = self.get_surface_thickness_mesh(wavefront)

        amp_factors = numpy.exp(-1.0 * att_coefficient * interpolated_profile / 2) # factor of 2 because it is amplitude
        phase_shifts = -1.0 * wavefront.get_wavenumber() * refraction_index_delta * interpolated_profile

        output_wavefront = wavefront.duplicate()
        output_wavefront.rescale_amplitudes(amp_factors.T)
        output_wavefront.add_phase_shifts(phase_shifts.T)

        return output_wavefront

    def to_python_code(self, data=None):
        txt  = ""
        txt += "\nfrom orangecontrib.esrf.wofry.widgets.extension.ow_thin_object_2d import WOThinObject"
        txt += "\n"
        if self.get_material() == "External":
            txt += "\noptical_element = WOThinObject(name='%s',file_with_thickness_mesh='%s',material='%s',refraction_index_delta=%g,att_coefficient=%g)" % \
                   (self.get_name(), self.get_file_with_thickness_mesh(), self.get_material(), self._refraction_index_delta, self._att_coefficient)
        else:
            txt += "\noptical_element = WOThinObject(name='%s',file_with_thickness_mesh='%s',material='%s')" % \
                   (self.get_name(), self.get_file_with_thickness_mesh(), self.get_material())
        txt += "\n"
        return txt


class WOThinObject1D(ThinObject, OpticalElementDecorator):
    def __init__(self,
                 name="Undefined",
                 file_with_thickness_mesh="",
                 material="",
                 refraction_index_delta=1e-07,
                 att_coefficient=0.0,
                 ):
        super().__init__(
                      name=name,
                      file_with_thickness_mesh=file_with_thickness_mesh,
                      material=material)

        self._refraction_index_delta = refraction_index_delta
        self._att_coefficient = att_coefficient

    def get_surface_thickness_mesh(self, wavefront):
        a = numpy.loadtxt(self.get_file_with_thickness_mesh())
        xx = a[:,0].copy()
        zz = a[:,1].copy()

        if zz.min() < 0: zz -= zz.min()
        x = wavefront.get_abscissas()
        interpolated_profile = numpy.interp(x, xx, zz)
        return x, interpolated_profile

    def get_refraction_index(self, photon_energy=10000.0):

        wave_length = codata.h * codata.c / codata.e / photon_energy

        if self.get_material() == "External": # Be
             return self._refraction_index_delta, \
                    self._att_coefficient

        if self.get_material() == "Be": # Be
            element = "Be"
            density = xraylib.ElementDensity(4)
        elif self.get_material() == "Al": # Al
            element = "Al"
            density = xraylib.ElementDensity(13)
        elif self.get_material() == "Diamond": # Diamond
            element = "C"
            density = 3.51
        else:
            raise Exception("Bad material: " + self.get_material())

        refraction_index = xraylib.Refractive_Index(element, photon_energy/1000, density)
        refraction_index_delta = 1 - refraction_index.real
        att_coefficient = 4*numpy.pi * (xraylib.Refractive_Index(element, photon_energy/1000, density)).imag / wave_length

        if False:
            print("\n\n\n ==========  parameters recovered from xraylib : ")
            print("Element: %s" % element)
            print("        density = %g " % density)
            print("Photon energy = %g eV" % (photon_energy))
            print("Refracion index delta = %g " % (refraction_index_delta))
            print("Attenuation coeff mu = %g m^-1" % (att_coefficient))

        return refraction_index_delta, att_coefficient

    def applyOpticalElement(self, wavefront, parameters=None, element_index=None):
        # return wavefront

        print("\n\n\n ==========  parameters from optical element : ")
        print(self.info())

        photon_energy = wavefront.get_photon_energy()

        refraction_index_delta, att_coefficient = self.get_refraction_index(photon_energy)

        x, interpolated_profile = self.get_surface_thickness_mesh(wavefront)
        #
        amp_factors = numpy.exp(-1.0 * att_coefficient * interpolated_profile / 2) # factor of 2 because it is amplitude
        phase_shifts = -1.0 * wavefront.get_wavenumber() * refraction_index_delta * interpolated_profile

        output_wavefront = wavefront.duplicate()
        output_wavefront.rescale_amplitudes(amp_factors)
        output_wavefront.add_phase_shifts(phase_shifts)

        return output_wavefront

    def to_python_code(self, data=None):
        txt  = ""
        txt += "\nfrom orangecontrib.esrf.wofry.util.thin_object import WOThinObject1D #TODO update"
        txt += "\n"
        if self.get_material() == "External":
            txt += "\noptical_element = WOThinObject1D(name='%s',file_with_thickness_mesh='%s',material='%s',refraction_index_delta=%g,att_coefficient=%g)" % \
                   (self.get_name(), self.get_file_with_thickness_mesh(), self.get_material(), self._refraction_index_delta, self._att_coefficient)
        else:
            txt += "\noptical_element = WOThinObject1D(name='%s',file_with_thickness_mesh='%s',material='%s')" % \
                   (self.get_name(), self.get_file_with_thickness_mesh(), self.get_material())

        txt += "\n"
        return txt

if __name__ == "__main__":

    import numpy
    from wofry.propagator.wavefront2D.generic_wavefront import GenericWavefront2D
    from wofry.propagator.wavefront1D.generic_wavefront import GenericWavefront1D
    from srxraylib.plot.gol import plot, plot_image


    #
    # 2D
    #
    if False:
        input_wavefront = GenericWavefront2D.initialize_wavefront_from_range(x_min=-0.0003, x_max=0.0003, y_min=-0.0003,
                                                                              y_max=0.0003, number_of_points=(400, 200))
        input_wavefront.set_photon_energy(10000)
        input_wavefront.set_plane_wave_from_complex_amplitude(complex_amplitude=complex(1, 0))

        optical_element = WOThinObject(name='ThinObject',
                                       file_with_thickness_mesh='/home/srio/Downloads/SRW_M_thk_res_workflow_a_FC_CDn01.dat.h5',
                                       material='Be')

        # no drift in this element
        output_wavefront = optical_element.applyOpticalElement(input_wavefront)

        #
        # ---- plots -----
        #
        plot_image(output_wavefront.get_intensity(), output_wavefront.get_coordinate_x(),
                   output_wavefront.get_coordinate_y(), aspect='auto', title='OPTICAL ELEMENT NR 1')


    #
    # 1D
    #

    if True:
        input_wavefront = GenericWavefront1D.initialize_wavefront_from_range(x_min=-0.0003, x_max=0.0003,
                                                                             number_of_points=400)
        input_wavefront.set_photon_energy(10000)
        input_wavefront.set_plane_wave_from_complex_amplitude(complex_amplitude=complex(1, 0))

        xx, yy, zz = read_surface_file('/home/srio/Downloads/SRW_M_thk_res_workflow_a_FC_CDn01.dat.h5')
        if zz.min() < 0: zz -= zz.min()
        nx, ny = zz.shape
        x = yy
        z = zz[nx//2,:]
        plot(x,z,title="x,z")

        f = open('profile.dat', 'w')
        for i in range(ny):
            f.write("%g %g\n" % (x[i], z[i]))
        f.close()
        print("File profile.dat written to disk.")

        optical_element = WOThinObject1D(name='ThinObject1D',
                                       file_with_thickness_mesh='profile.dat',
                                       material='Be')


        print(optical_element.info())

        # no drift in this element
        output_wavefront = optical_element.applyOpticalElement(input_wavefront)

        #
        # ---- plots -----
        #
        plot(output_wavefront.get_abscissas(), output_wavefront.get_intensity(),
                   title='OPTICAL ELEMENT NR 1')


import os, sys
import numpy

from PyQt5.QtCore import QRect, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox, QScrollArea, QTableWidget, QTableWidgetItem, QHeaderView, QAbstractItemView, QWidget, QLabel, QSizePolicy
from PyQt5.QtGui import QTextCursor,QFont, QPalette, QColor, QPainter, QBrush, QPen, QPixmap

import orangecanvas.resources as resources

from orangewidget import gui, widget
from orangewidget.settings import Setting

from oasys.widgets.widget import OWWidget
from oasys.widgets import gui as oasysgui

# from orangecontrib.shadow.util.shadow_objects import ConicCoefficientsPreProcessorData # TODO to be added
from orangecontrib.esrf.shadow.util.shadow_objects import ConicCoefficientsPreProcessorData

from oasys.util.oasys_util import EmittingStream

# from scipy.optimize import fsolve


#
#  TODO: correct rays when separation!=0. Need to calculate intersection of hyperbola and straight line
#  from (x_pmin,y_pmin) to (0,0)
#
class OWWolterCenteredCalculator(OWWidget):
    name = "Wolter-Centered Calculator"
    id = "WolterCenteredCalculator"
    description = "Calculation of coefficients for Wolter-centered systems"
    icon = "icons/wolter2.png"
    author = "Manuel Sanchez del Rio"
    maintainer_email = "srio@esrf.eu"
    priority = 6
    category = ""
    keywords = ["oasys", "wolter", "telescope", "advanced KB"]

    outputs = [{"name":"ConicCoeff_1_PreProcessor_Data",
                "type":ConicCoefficientsPreProcessorData,
                "doc":"ConicCoeff #1 PreProcessor Data",
                "id":"ConicCoeff_1_PreProcessor_Data"},
               {"name": "ConicCoeff_2_PreProcessor_Data",
                "type": ConicCoefficientsPreProcessorData,
                "doc": "ConicCoeff #2 PreProcessor Data",
                "id": "ConicCoeff_2_PreProcessor_Data"},
               ]

    want_main_area = True
    want_control_area = True

    MAX_WIDTH = 1320
    MAX_HEIGHT = 700

    IMAGE_WIDTH = 860
    IMAGE_HEIGHT = 645

    CONTROL_AREA_WIDTH = 405
    TABS_AREA_HEIGHT = 650 #18

    #################

    # setup_type = Setting(0)

    p1 = Setting(-0.00194644)
    p2 = Setting(1.904995)

    theta1 = Setting(0.0159872)

    separation = Setting(0.0)
    ellipse_flag = Setting(0)
    ellipse_2c = Setting(10.0)
    npoints = Setting(400)
    tab=[]

    usage_path = os.path.join(resources.package_dirname("orangecontrib.esrf.shadow.widgets.extension"), "icons", "use_wolter.png")

    def __init__(self):
        super().__init__()

        self.runaction = widget.OWAction("Calculate", self)
        self.runaction.triggered.connect(self.calculate)
        self.addAction(self.runaction)


        geom = QApplication.desktop().availableGeometry()
        self.setGeometry(QRect(round(geom.width() * 0.05),
                               round(geom.height() * 0.05),
                               round(min(geom.width() * 0.98, self.MAX_WIDTH)),
                               round(min(geom.height() * 0.95, self.MAX_HEIGHT))))

        self.setMaximumHeight(self.geometry().height())
        self.setMaximumWidth(self.geometry().width())

        gui.separator(self.controlArea)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_calc = oasysgui.createTabPage(tabs_setting, "Calculate")
        # tab_out = oasysgui.createTabPage(tabs_setting, "Plot")
        tab_usa = oasysgui.createTabPage(tabs_setting, "Use of the Widget")


        #
        #-------------------- calculate
        #

        button = gui.button(tab_calc, self, "Calculate", callback=self.calculate)

        tab_step_1 = oasysgui.widgetBox(tab_calc, "Calculation Parameters", addSpace=True, orientation="vertical", height=600)

        box = oasysgui.widgetBox(tab_step_1, "Main inputs", orientation="vertical")

        # gui.comboBox(box, self, "setup_type", label="Setup type", labelWidth=260,
        #              items=["Wolter-I",
        #                     # "Wolter-II",
        #                     # "Wolter-III",
        #                     ],
        #              callback=self.update_panel, sendSelectedValue=False, orientation="horizontal")

        self.w_p1 = oasysgui.lineEdit(box, self, "p1", "parabola directrix coord (z=-p=-f/2)",
                            labelWidth=240, valueType=float, orientation="horizontal", tooltip="p1")
        self.w_p2 = oasysgui.lineEdit(box, self, "p2", "hyp interfocal dist 2c (>=0 W1, <=0 W2) [m]", labelWidth=240, valueType=float, orientation="horizontal")
        self.w_theta1 = oasysgui.lineEdit(box, self, "theta1", "Grazing angle at mirrors intersection [rad]",
                            labelWidth=260, valueType=float, orientation="horizontal", tooltip="theta1",
                            callback=self.update_panel)
        oasysgui.lineEdit(box, self, "separation", "separation (<=0 W1, >=0 W2)[m]", labelWidth=260,
                                      valueType=float, orientation="horizontal")

        box = oasysgui.widgetBox(tab_step_1, "Ellipse inputs", orientation="vertical")
        gui.comboBox(box, self, "ellipse_flag", label="Replace parabola by ellipse?", labelWidth=260,
                     items=["No","Yes"], tooltip="ellipse_flag",
                     callback=self.update_panel, sendSelectedValue=False, orientation="horizontal")
        self.w_ellipse_2c = oasysgui.widgetBox(box, "", orientation="vertical")
        oasysgui.lineEdit(self.w_ellipse_2c, self, "ellipse_2c", "Ellipse focus (z=2c)", labelWidth=260,
                    valueType=float, orientation="horizontal", callback=self.update_panel, tooltip="ellipse_2c")

        box = oasysgui.widgetBox(tab_step_1, "Other inputs", orientation="vertical")
        oasysgui.lineEdit(box, self, "npoints", "Points (for plot)", labelWidth=260, valueType=int,
                          orientation="horizontal", callback=self.update_panel, tooltip="npoints")

        #
        #-------------------- Use
        #

        tab_usa.setStyleSheet("background-color: white;")

        usage_box = oasysgui.widgetBox(tab_usa, "", addSpace=True, orientation="horizontal")

        label = QLabel("")
        label.setAlignment(Qt.AlignCenter)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        label.setPixmap(QPixmap(self.usage_path))

        usage_box.layout().addWidget(label)

        #
        #
        #
        gui.rubber(self.controlArea)

        self.initializeTabs()

        gui.rubber(self.mainArea)

        self.update_panel()

    def update_panel(self):
        self.w_p1.setVisible(True)
        self.w_p2.setVisible(True)
        self.w_theta1.setVisible  (True)
        if self.ellipse_flag:
            self.w_ellipse_2c.setVisible(True)
        else:
            self.w_ellipse_2c.setVisible(False)

    def calculate(self):
        try:
            self.shadow_output.setText("")
            self.design_output.setText("")

            sys.stdout = EmittingStream(textWritten=self.writeStdOut)

            self.check_fields()

            results_txt = ""
            print("####################################################")
            print("# DESIGN PHASE")
            print("####################################################\n")

            tkt = self.wolter1_centered()

            p = tkt['p']
            a_h = tkt['a_h']
            b_h = tkt['b_h']
            c_h = tkt['c_h']
            a_e = tkt['a_e']
            b_e = tkt['b_e']
            c_e = tkt['c_e']
            x_pmin = tkt['x_pmin']
            y_pmin = tkt['y_pmin']
            x_he = tkt['x_he']
            y_he = tkt['y_he']

            results_txt += "\nparabola p=%f"  % p
            if self.ellipse_flag:
                results_txt += "\nellipse (replacing parabola) a=%f" % a_e
                results_txt += "\nellipse (replacing parabola) b=%f" % b_e
                results_txt += "\nellipse (replacing parabola) c=%f" % c_e
            results_txt += "\nhyperbola a=%f" % a_h
            results_txt += "\nhyperbola b=%f" % b_h
            results_txt += "\nhyperbola c=%f" % c_h


            results_txt += "\n\n\n    oe1(normalized)      oe2(normalized)"

            if self.ellipse_flag:
                for i in range(10):
                    results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i,
                            tkt['ccc3'][i]/tkt['ccc3'][0], tkt['ccc2'][i]/tkt['ccc2'][0])

                results_txt += "\n\n\n    oe1           oe2 "
                for i in range(10):
                    results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i, tkt['ccc3'][i], tkt['ccc2'][i])


            else:
                for i in range(10):
                    results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i,
                            tkt['ccc1'][i]/tkt['ccc1'][0], tkt['ccc2'][i]/tkt['ccc2'][0])

                results_txt += "\n\n\n    oe1           oe2 "
                for i in range(10):
                    results_txt += "\nccc[%d]       %10.4g       %10.4g  " % (i, tkt['ccc1'][i], tkt['ccc2'][i])


            self.design_output.setText(results_txt)

            if self.ellipse_flag:
                Pe = numpy.sqrt((2 * c_e) ** 2 + 0)
                Qe = numpy.sqrt(x_he ** 2 + y_he ** 2)
                Fe = 1 / (1 / Pe + 1 / Qe)
                print("\n\nEllipse p = ", Pe)
                print("Ellipse q = ", Qe)
                print("Ellipse f = ", Fe)

                Ph = Qe
                Qh = numpy.sqrt((x_pmin - 2 * c_h) ** 2 + y_pmin ** 2)
                Fh = 1 / (1 / Ph + 1 / Qh)
                print("\nHyperbola p = ", Ph)
                print("Hyperbola q = ", Qh)
                print("Hyperbola f = ", Fh)

                F = 1. / (1 / Fe + 1 / Fh)
                print("\nCombined f = ", F)

                print("N.A. (at cross point)", numpy.sin(y_he/(2*c_e)))
                print("Angle (at cross point)", y_he / (2 * c_e))

            else:
                Pp = None
                Qp = numpy.sqrt(x_pmin**2 + y_pmin**2)
                Fp = Qp
                print("\n\nParabola p = NOT DEFINED ")
                print("Parabola q = ", Qp)
                print("Parabola f = ", Qp)

                Ph = Qp
                Qh = numpy.sqrt((x_pmin-2*c_h)**2 + y_pmin**2)
                Fh = 1/(1/Ph+1/Qh)
                print("\nHyperbola p = ", Ph)
                print("Hyperbola q = ", Qh)
                print("Hyperbola f = ", Fh)

                F = 1. / (1 / Fp + 1 / Fh)
                print("\nCombined f = ", F)



            #
            # plot data
            #
            self.progressBarInit()
            x = numpy.linspace(x_pmin*1.5, -x_pmin*1.5/5, self.npoints)

            #
            # parabola
            #
            y1a =  numpy.sqrt(2 * p * x + p ** 2)
            y1b = -numpy.sqrt(2 * p * x + p ** 2)
            theta_p = numpy.arctan((2 * x / p + 1) ** (-1 / 2))

            #
            # hyperbola
            #
            y2a =  b_h * numpy.sqrt(((x - c_h) / a_h) ** 2 - 1)
            y2b = -b_h * numpy.sqrt(((x - c_h) / a_h) ** 2 - 1)
            p_x =  numpy.sqrt((x - 2 * c_h) ** 2 + y2a ** 2)
            q_h_x =  numpy.sqrt(x ** 2 + y2a ** 2)
            theta_h = numpy.arcsin(b_h/numpy.sqrt(p_x*q_h_x))

            #
            # ellipse
            #
            y3a =   b_e * numpy.sqrt(1 - ((x - c_e) / a_e) ** 2)
            y3b =  -b_e * numpy.sqrt(1 - ((x - c_e) / a_e) ** 2)
            p_e_x = numpy.sqrt((x - 2 * c_e) ** 2 + y3a ** 2)
            q_e_x =  numpy.sqrt(x ** 2 + y3a ** 2)
            theta_e = numpy.arcsin(b_e/numpy.sqrt(p_e_x*q_e_x))

            # plot oe 1

            if self.ellipse_flag:
                self.plot_multi_data1D([-x, -x, -x, -x], [y1a, y1b, y3a, y3b],
                                       10, 2, 0,
                                       title="oe1 (parabola or ellipse)", xtitle="-z [m] (along optical axis)", ytitle="x,y [m]",
                                       ytitles=["parabola+", "parabola-","ellipse+", "ellipse-"],
                                       colors=['blue', 'blue','green', 'green'],
                                       replace=True,
                                       control=False,
                                       xrange=None,
                                       yrange=None,
                                       symbol=['', '','', ''])

            else:
                self.plot_multi_data1D([-x,-x], [y1a,y1b],
                                      10, 2, 0,
                                      title="parabola", xtitle="-z [m] (along optical axis)", ytitle="x,y [m]",
                                      ytitles=["parabola+","parabola-"],
                                      colors=['blue','blue'],
                                      replace=True,
                                      control=False,
                                      xrange=None,
                                      yrange=None,
                                      symbol=['',''])

            # plot oe2
            self.plot_multi_data1D([-x,-x], [y2a,y2b],
                                  20, 3, 1,
                                  title="hyperbola", xtitle="-z [m] (along optical axis)", ytitle="x,y [m]",
                                  ytitles=["hyperbola+","hyperbola-"],
                                  colors=['red','red'],
                                  replace=True,
                                  control=False,
                                  xrange=None,
                                  yrange=None,
                                  symbol=['',''])

            #
            # plot joint oe1+oe2
            #


            x_c = numpy.array([x_pmin*1.5, x_pmin, 2*c_h, 2*c_h,  x_pmin,  x_pmin*1.5])
            y_c = numpy.array([y_pmin,     y_pmin, 0  , 0  , -y_pmin, -y_pmin    ])

            x_c2 = numpy.array([2*c_e, x_he, 2*c_h, 2*c_h,  x_he,  2*c_e])
            y_c2 = numpy.array([0,     y_he, 0    ,   0  , -y_he,  0    ])

            x_c3 = numpy.array([x_pmin, 0])
            y_c3 = numpy.array([y_pmin, 0])
            if self.ellipse_flag:
                x_c3 = numpy.array([x_he, 0,  x_he])
                y_c3 = numpy.array([y_he, 0, -y_he])
                self.plot_multi_data1D([-x,-x,-x,-x, -x_c, -x, -x, -x_c2, -x_c3], [y1a,y1b,y2a,y2b,y_c, y3a, y3b, y_c2, y_c3],
                                      80, 4, 2,
                                      title="parabola+hyperbola+ellipse", xtitle="-z [m] (along optical axis)", ytitle="x,y [m]",
                                      ytitles=["parabola+","parabola-","hyperbola+","hyperbola-", "ray at par+hyp crossing", "ellipse+","ellipse-", "ray at ell+hyp crossing", "ray to common focus"],
                                      colors=['blue','blue','red','red','k','green','green','k','pink'],
                                      replace=True,
                                      control=False,
                                      xrange=[-x.min(),- x.max()],
                                      yrange=None,
                                      symbol=['','','','','','','','',''],)
            else:
                x_c3 = numpy.array([x_pmin, 0,  x_pmin])
                y_c3 = numpy.array([y_pmin, 0, -y_pmin])
                self.plot_multi_data1D([-x,-x,-x,-x, -x_c, -x_c3], [y1a,y1b,y2a,y2b,y_c,y_c3],
                                      80, 4, 2,
                                      title="parabola+hyperbola", xtitle="-z [m] (along optical axis)", ytitle="x,y [m]",
                                      ytitles=["parabola+","parabola-","hyperbola+","hyperbola-", "ray at par+hyp crossing", "ray to common focus"],
                                      colors=['blue','blue','red','red','k','pink'],
                                      replace=True,
                                      control=False,
                                      xrange=None,
                                      yrange=None,
                                      symbol=['','','','','',''])



            # plot angles
            XXhe = numpy.array([x_pmin, x_pmin])
            YYhe = numpy.array([0, numpy.nanmax(theta_h)])
            if self.ellipse_flag:
                Xhe = numpy.array([x_he,x_he])
                Yhe = numpy.array([0,numpy.nanmax(theta_h)])
                self.plot_multi_data1D([-x,-x,-x,-XXhe,-x,-Xhe],
                                       [1e3*(x*0+1)*self.theta1,
                                        1e3*theta_p,
                                        1e3*theta_h,
                                        1e3 * YYhe,
                                        1e3*theta_e,
                                        1e3*Yhe,
                                        ],
                                      90, 5, 3,
                                      title="Grazing incident angles", xtitle="-z [m] (along optical axis)", ytitle="angle [mrad]",
                                      ytitles=["design","parabola","hyperbola",'par+hyp crossing',"ellipse",'ell+hyp crossing'],
                                      colors=['black','blue','red','k','green','pink'],
                                      replace=True,
                                      control=False,
                                      xrange=None,
                                      yrange=None,
                                      symbol=['','','','','',''])
            else:

                self.plot_multi_data1D([-x,-x,-x,-XXhe],
                                       [1e3*(x*0+1)*self.theta1,
                                        1e3*theta_p,
                                        1e3*theta_h,
                                        1e3 * YYhe,],
                                      90, 5, 3,
                                      title="Grazing incident angles", xtitle="-z [m] (along optical axis)", ytitle="angle [mrad]",
                                      ytitles=["design","parabola","hyperbola",'par+hyp crossing'],
                                      colors=['black','blue','red','k'],
                                      replace=True,
                                      control=False,
                                      xrange=None,
                                      yrange=None,
                                      symbol=['','','',''])

            #
            # send data
            #
            print("\n\n\n\n")
            print("####################################################")
            print("# RAY-TRACING PHASE")
            print("####################################################\n")

            if self.ellipse_flag:
                preprocessor_oe1 = ConicCoefficientsPreProcessorData(
                    conic_coefficient_0=tkt['ccc3'][0],
                    conic_coefficient_1=tkt['ccc3'][1],
                    conic_coefficient_2=tkt['ccc3'][2],
                    conic_coefficient_3=tkt['ccc3'][3],
                    conic_coefficient_4=tkt['ccc3'][4],
                    conic_coefficient_5=tkt['ccc3'][5],
                    conic_coefficient_6=tkt['ccc3'][6],
                    conic_coefficient_7=tkt['ccc3'][7],
                    conic_coefficient_8=tkt['ccc3'][8],
                    conic_coefficient_9=tkt['ccc3'][9],
                    source_plane_distance=self.ellipse_2c,
                    image_plane_distance=-tkt['x_he'],
                    angles_respect_to=0,
                    incidence_angle_deg=0.0,
                    reflection_angle_deg=0.0,
                    mirror_orientation_angle=2,  # that means 180 deg
                    title="Primary mirror: Elliptical",
                )

                preprocessor_oe2 = ConicCoefficientsPreProcessorData(
                    conic_coefficient_0= tkt['ccc2'][0],
                    conic_coefficient_1= tkt['ccc2'][1],
                    conic_coefficient_2= tkt['ccc2'][2],
                    conic_coefficient_3= tkt['ccc2'][3],
                    conic_coefficient_4= tkt['ccc2'][4],
                    conic_coefficient_5= tkt['ccc2'][5],
                    conic_coefficient_6= tkt['ccc2'][6],
                    conic_coefficient_7= tkt['ccc2'][7],
                    conic_coefficient_8= tkt['ccc2'][8],
                    conic_coefficient_9= tkt['ccc2'][9],
                    source_plane_distance=tkt['x_he'],
                    image_plane_distance=2*tkt['c_h'],
                    angles_respect_to=0,
                    incidence_angle_deg=0.0,
                    reflection_angle_deg=0.0,
                    mirror_orientation_angle=0,
                    title="Secondary mirror: Hyperbolic",
                    )
            else:
                preprocessor_oe1 = ConicCoefficientsPreProcessorData(
                    conic_coefficient_0 = tkt['ccc1'][0],
                    conic_coefficient_1 = tkt['ccc1'][1],
                    conic_coefficient_2 = tkt['ccc1'][2],
                    conic_coefficient_3 = tkt['ccc1'][3],
                    conic_coefficient_4 = tkt['ccc1'][4],
                    conic_coefficient_5 = tkt['ccc1'][5],
                    conic_coefficient_6 = tkt['ccc1'][6],
                    conic_coefficient_7 = tkt['ccc1'][7],
                    conic_coefficient_8 = tkt['ccc1'][8],
                    conic_coefficient_9 = tkt['ccc1'][9],
                    source_plane_distance=None,
                    image_plane_distance=-tkt['x_pmin'],
                    angles_respect_to=0,
                    incidence_angle_deg=0.0,
                    reflection_angle_deg=0.0,
                    mirror_orientation_angle=2, # that means 180 deg
                    title="Primary mirror: Parabolic",
                    )

                preprocessor_oe2 = ConicCoefficientsPreProcessorData(
                    conic_coefficient_0= tkt['ccc2'][0],
                    conic_coefficient_1= tkt['ccc2'][1],
                    conic_coefficient_2= tkt['ccc2'][2],
                    conic_coefficient_3= tkt['ccc2'][3],
                    conic_coefficient_4= tkt['ccc2'][4],
                    conic_coefficient_5= tkt['ccc2'][5],
                    conic_coefficient_6= tkt['ccc2'][6],
                    conic_coefficient_7= tkt['ccc2'][7],
                    conic_coefficient_8= tkt['ccc2'][8],
                    conic_coefficient_9= tkt['ccc2'][9],
                    source_plane_distance=tkt['x_pmin'],
                    image_plane_distance=2*tkt['c_h'],
                    angles_respect_to=0,
                    incidence_angle_deg=0.0,
                    reflection_angle_deg=0.0,
                    mirror_orientation_angle=0,
                    title="Secondary mirror: Hyperbolic",
                    )

            for obj in [preprocessor_oe1,preprocessor_oe2]:
                print("\nChanging/setting parameters of shadow conic coefficient mirror: ", obj.title)
                print("conic_coefficient_0",      obj.conic_coefficient_0)
                print("conic_coefficient_1",      obj.conic_coefficient_1)
                print("conic_coefficient_2",      obj.conic_coefficient_2)
                print("conic_coefficient_3",      obj.conic_coefficient_3)
                print("conic_coefficient_4",      obj.conic_coefficient_4)
                print("conic_coefficient_5",      obj.conic_coefficient_5)
                print("conic_coefficient_6",      obj.conic_coefficient_6)
                print("conic_coefficient_7",      obj.conic_coefficient_7)
                print("conic_coefficient_8",      obj.conic_coefficient_8)
                print("conic_coefficient_9",      obj.conic_coefficient_9)
                if obj.source_plane_distance    is not None: print("source_plane_distance",    obj.source_plane_distance)
                if obj.image_plane_distance     is not None: print("image_plane_distance",     obj.image_plane_distance)
                if obj.angles_respect_to        is not None: print("angles_respect_to",        obj.angles_respect_to)
                if obj.incidence_angle_deg      is not None: print("incidence_angle_deg",      obj.incidence_angle_deg)
                if obj.reflection_angle_deg     is not None: print("reflection_angle_deg",     obj.reflection_angle_deg)
                if obj.mirror_orientation_angle is not None: print("mirror_orientation_angle", obj.mirror_orientation_angle)


            self.send("ConicCoeff_1_PreProcessor_Data", preprocessor_oe1)
            self.send("ConicCoeff_2_PreProcessor_Data", preprocessor_oe2)

        except Exception as exception:
            QMessageBox.critical(self, "Error",
                                 str(exception),
                                 QMessageBox.Ok)

        self.progressBarFinished()


    def initializeTabs(self):
        self.tabs = oasysgui.tabWidget(self.mainArea)

        self.tab = [oasysgui.createTabPage(self.tabs, "Design parameters"),
                    oasysgui.createTabPage(self.tabs, "Output"),
                    oasysgui.createTabPage(self.tabs, "oe1 Profile"),
                    oasysgui.createTabPage(self.tabs, "oe2 Profile"),
                    oasysgui.createTabPage(self.tabs, "join profile"),
                    oasysgui.createTabPage(self.tabs, "incident angle"),
        ]

        for tab in self.tab:
            tab.setFixedHeight(self.IMAGE_HEIGHT)
            tab.setFixedWidth(self.IMAGE_WIDTH)


        self.design_output = oasysgui.textArea()
        tmp2 = oasysgui.widgetBox(self.tab[0], "Design output", addSpace=True, orientation="horizontal", height = self.IMAGE_HEIGHT-4) #, width=410)
        tmp2.layout().addWidget(self.design_output)

        self.shadow_output = oasysgui.textArea() #height=self.IMAGE_HEIGHT-5, width=400)
        tmp1 = oasysgui.widgetBox(self.tab[1], "System output", addSpace=True, orientation="horizontal", height = self.IMAGE_HEIGHT-4) #, width=410)
        tmp1.layout().addWidget(self.shadow_output)

        #
        self.plot_canvas = [None, None, None, None]

        self.plot_canvas[0] = oasysgui.plotWindow(roi=False, control=False, position=True)
        self.plot_canvas[0].setDefaultPlotLines(True)
        # self.plot_canvas[0].setActiveCurveColor(color='blue')
        self.plot_canvas[0].setGraphYLabel("Z [nm]")
        self.plot_canvas[0].setGraphTitle("oe1 Profile")
        self.plot_canvas[0].setInteractiveMode(mode='zoom')

        self.plot_canvas[1] = oasysgui.plotWindow(roi=False, control=False, position=True)
        self.plot_canvas[1].setDefaultPlotLines(True)
        # self.plot_canvas[1].setActiveCurveColor(color='red')
        self.plot_canvas[1].setGraphYLabel("Z [nm]")
        self.plot_canvas[1].setGraphTitle("oe2 Profile")
        self.plot_canvas[1].setInteractiveMode(mode='zoom')

        self.plot_canvas[2] = oasysgui.plotWindow(roi=False, control=False, position=True)
        self.plot_canvas[2].setDefaultPlotLines(True)
        # self.plot_canvas[2].setActiveCurveColor(color='blue')
        self.plot_canvas[2].setGraphYLabel("Z [nm]")
        self.plot_canvas[2].setGraphTitle("Joint Profile")
        self.plot_canvas[2].setInteractiveMode(mode='zoom')

        self.plot_canvas[3] = oasysgui.plotWindow(roi=False, control=False, position=True)
        self.plot_canvas[3].setDefaultPlotLines(True)
        # self.plot_canvas[3].setActiveCurveColor(color='black')
        self.plot_canvas[3].setGraphYLabel("grazing angle [mrad]")
        self.plot_canvas[3].setGraphTitle("Grazing incident angle")
        self.plot_canvas[3].setInteractiveMode(mode='zoom')

        self.tab[2].layout().addWidget(self.plot_canvas[0])
        self.tab[3].layout().addWidget(self.plot_canvas[1])
        self.tab[4].layout().addWidget(self.plot_canvas[2])
        self.tab[5].layout().addWidget(self.plot_canvas[3])

        self.tabs.setCurrentIndex(0)

    def check_fields(self):
        pass
        # if self.setup_type > 0:
        #     raise Exception(NotImplementedError)
        # else:
        #     pass

        # self.dimension_x = congruence.checkStrictlyPositiveNumber(self.dimension_x, "Dimension X")
        # self.step_x = congruence.checkStrictlyPositiveNumber(self.step_x, "Step X")
        #
        # congruence.checkLessOrEqualThan(self.step_x, self.dimension_x/2, "Step Width", "Width/2")
        #
        # if self.modify_y == 1 or self.modify_y == 2:
        #     self.new_length = congruence.checkStrictlyPositiveNumber(self.new_length, "New Length")
        #
        # if self.renormalize_y == 1:
        #     self.rms_y = congruence.checkPositiveNumber(self.rms_y, "Rms Y")
        #
        # congruence.checkDir(self.heigth_profile_file_name)

    def writeStdOut(self, text):
        cursor = self.shadow_output.textCursor()
        cursor.movePosition(QTextCursor.End)
        cursor.insertText(text)
        self.shadow_output.setTextCursor(cursor)
        self.shadow_output.ensureCursorVisible()

    # def plot_data1D(self, x, y, progressBarValue, tabs_canvas_index, plot_canvas_index, title="", xtitle="", ytitle="",
    #                 log_x=False, log_y=False,
    #                 color='blue',
    #                 replace=True, control=False, calculate_fwhm=True,
    #                 xrange=None, yrange=None, symbol=''):
    #
    #     if tabs_canvas_index is None: tabs_canvas_index = 0 #back compatibility?
    #
    #
    #     self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(0))
    #
    #     self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(parent=None,
    #                                                               backend=None,
    #                                                               resetzoom=True,
    #                                                               autoScale=False,
    #                                                               logScale=True,
    #                                                               grid=True,
    #                                                               curveStyle=True,
    #                                                               colormap=False,
    #                                                               aspectRatio=False,
    #                                                               yInverted=False,
    #                                                               copy=True,
    #                                                               save=True,
    #                                                               print_=True,
    #                                                               control=control,
    #                                                               position=True,
    #                                                               roi=False,
    #                                                               mask=False,
    #                                                               fit=False)
    #
    #
    #     self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
    #     self.plot_canvas[plot_canvas_index].setActiveCurveColor(color='blue')
    #     self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
    #     self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)
    #
    #     # ALLOW FIT BUTTON HERE
    #     self.plot_canvas[plot_canvas_index].fitAction.setVisible(True)
    #
    #     # overwrite FWHM and peak values
    #     if calculate_fwhm:
    #         try:
    #             t = numpy.where(y>=max(y)*0.5)
    #             x_left,x_right =  x[t[0][0]], x[t[0][-1]]
    #
    #             self.plot_canvas[plot_canvas_index].addMarker(x_left, 0.5*y.max(), legend="G1",
    #                                                           text="FWHM=%5.2f"%(numpy.abs(x_right-x_left)),
    #                                                           color="pink",selectable=False, draggable=False,
    #                                                           symbol="+", constraint=None)
    #             self.plot_canvas[plot_canvas_index].addMarker(x_right, 0.5*y.max(), legend="G2", text=None, color="pink",
    #                                                           selectable=False, draggable=False, symbol="+", constraint=None)
    #         except:
    #             pass
    #
    #     self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])
    #
    #     self.plot_histo(self.plot_canvas[plot_canvas_index], x, y, title, xtitle, ytitle, color, replace, symbol=symbol)
    #
    #     self.plot_canvas[plot_canvas_index].setXAxisLogarithmic(log_x)
    #     self.plot_canvas[plot_canvas_index].setYAxisLogarithmic(log_y)
    #
    #
    #     if xrange is not None:
    #         self.plot_canvas[plot_canvas_index].setGraphXLimits(xrange[0],xrange[1])
    #     if yrange is not None:
    #         self.plot_canvas[plot_canvas_index].setGraphYLimits(yrange[0],yrange[1])
    #
    #     if min(y) < 0:
    #         if log_y:
    #             self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y)*1.2, max(y)*1.2)
    #         else:
    #             self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y)*1.01, max(y)*1.01)
    #     else:
    #         if log_y:
    #             self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y), max(y)*1.2)
    #         else:
    #             self.plot_canvas[plot_canvas_index].setGraphYLimits(min(y)*0.99, max(y)*1.01)
    #
    #     self.progressBarSet(progressBarValue)


    def plot_multi_data1D(self, x_list, y_list,
                    progressBarValue, tabs_canvas_index, plot_canvas_index,
                    title="", xtitle="",
                    ytitle="",
                    ytitles= [""],
                    colors = ['green'],
                    replace=True,
                    control=False,
                    xrange=None,
                    yrange=None,
                    symbol=['']):

        if len(y_list) != len(ytitles):
            ytitles = ytitles * len(y_list)

        if len(y_list) != len(colors):
            colors = colors * len(y_list)
        if len(y_list) != len(symbol):
            symbols = symbol * len(y_list)
        else:
            symbols = symbol

        if tabs_canvas_index is None: tabs_canvas_index = 0 #back compatibility?

        self.tab[tabs_canvas_index].layout().removeItem(self.tab[tabs_canvas_index].layout().itemAt(0))

        self.plot_canvas[plot_canvas_index] = oasysgui.plotWindow(parent=None,
                                                                  backend=None,
                                                                  resetzoom=True,
                                                                  autoScale=False,
                                                                  logScale=True,
                                                                  grid=True,
                                                                  curveStyle=True,
                                                                  colormap=False,
                                                                  aspectRatio=False,
                                                                  yInverted=False,
                                                                  copy=True,
                                                                  save=True,
                                                                  print_=True,
                                                                  control=control,
                                                                  position=True,
                                                                  roi=False,
                                                                  mask=False,
                                                                  fit=False)


        self.plot_canvas[plot_canvas_index].setDefaultPlotLines(True)
        self.plot_canvas[plot_canvas_index].setActiveCurveColor(color=colors[0])
        self.plot_canvas[plot_canvas_index].setGraphXLabel(xtitle)
        self.plot_canvas[plot_canvas_index].setGraphYLabel(ytitle)

        self.tab[tabs_canvas_index].layout().addWidget(self.plot_canvas[plot_canvas_index])


        for i in range(len(y_list)):
            # print(">>>>>>>>>>>>>>>>>>>> ADDING PLOT INDEX", i, x_list[i].shape, y_list[i].shape,ytitles[i],symbols[i],colors[i])
            self.plot_canvas[plot_canvas_index].addCurve(x_list[i], y_list[i],
                                         ytitles[i],
                                         xlabel=xtitle,
                                         ylabel=ytitle,
                                         symbol=symbols[i],
                                         color=colors[i])
        #
        self.plot_canvas[plot_canvas_index].getLegendsDockWidget().setFixedHeight(150)
        self.plot_canvas[plot_canvas_index].getLegendsDockWidget().setVisible(True)
        self.plot_canvas[plot_canvas_index].setActiveCurve(ytitles[0])
        self.plot_canvas[plot_canvas_index].replot()


        if xrange is not None:
            self.plot_canvas[plot_canvas_index].setGraphXLimits(xrange[0],xrange[1])
        if yrange is not None:
            self.plot_canvas[plot_canvas_index].setGraphYLimits(yrange[0],yrange[1])

        # if numpy.amin(numpy.array(y_list)) < 0:
        #     self.plot_canvas[plot_canvas_index].setGraphYLimits(numpy.amin(numpy.array(y_list))*1.01, numpy.amax(numpy.array(y_list))*1.01)
        # else:
        #     self.plot_canvas[plot_canvas_index].setGraphYLimits(numpy.amin(numpy.array(y_list))*0.99, numpy.amax(numpy.array(y_list))*1.01)

        self.progressBarSet(progressBarValue)


    # def plot_data2D(self, data2D, dataX, dataY, progressBarValue, tabs_canvas_index, plot_canvas_index,
    #                 title="",xtitle="", ytitle=""):
    #
    #     if self.view_type == 0:
    #         pass
    #     elif self.view_type == 1:
    #         self.plot_data2D_only_image(data2D, dataX, dataY, progressBarValue, tabs_canvas_index,plot_canvas_index,
    #                      title=title, xtitle=xtitle, ytitle=ytitle)
    #     elif self.view_type == 2:
    #         self.plot_data2D_with_histograms(data2D, dataX, dataY, progressBarValue, tabs_canvas_index,plot_canvas_index,
    #                      title=title, xtitle=xtitle, ytitle=ytitle)

    def wolter1_centered(self, verbose = True):
        #  centered system parabola-hyperbola
        # f11 = -0.00194644,
        # f12 = 0.0,
        # f21 = 1.905,
        # f22 = 0.0,
        # theta = 0.0159872,
        # verbose = 1,

        f11 = self.p1
        f12 = 0.0
        f21 = self.p2
        f22 = 0.0
        theta = self.theta1
        c_e = self.ellipse_2c / 2

        if f12 != 0.0 or f22 != 0:
            raise Exception("Is your origin at the common focus?")

        # p = numpy.abs(f11-f12)
        p = f12 - f11 # sign is important, f11 < 0, f21 > 0

        # intersection point at the parabola matching angle (https://doi.org/10.1107/S1600577522004593)
        x_pmin = (p / 2) / (numpy.tan(theta)) ** 2 - (p/2)
        y_pmin = numpy.sqrt( 2 * p * x_pmin + p**2)

        #
        # hyperbola
        #

        c_h = f21/2

        # get a from the hyperbola
        # (x-c)/a)^2 - (y/b)^2 = 1
        # b^2 = c^2 - a^2
        x_pmin1 = x_pmin + self.separation
        y_pmin1 = numpy.sqrt( 2 * p * x_pmin1 + p**2)
        A = 1
        B = -x_pmin1**2 + 2 * c_h * x_pmin1 - 2 * c_h**2 - y_pmin1**2
        C = c_h**4 + c_h**2 * x_pmin1**2 - 2 * c_h**3 * x_pmin1
        S1 = numpy.sqrt( (-B + numpy.sqrt(B**2 - 4*A*C)) / (2 * A) )
        S2 = numpy.sqrt( (-B - numpy.sqrt(B ** 2 - 4 * A * C)) / (2 * A) )
        a_h = numpy.min((S1,S2))
        b_h = numpy.sqrt(c_h**2 - a_h**2)
        # b_h2 = numpy.sqrt(numpy.sqrt((x_pmin-c_h) ** 2 + y_pmin ** 2) * numpy.sqrt(x_pmin ** 2 + y_pmin ** 2)) * numpy.sin(theta)


        # coeffs

        # Parabola
        # y^2 = p(2 * x + p) (Underwood)
        # x^2 + y^2 = 2 p z + p^2 (Shadow)
        ccc_centered_parabola = numpy.array([1, 1, 0, 0, 0, 0, 0, 0, -2 * p, -p ** 2])

        # Hyperbola
        # (x-c)^2/a^2 - y^2/b^2 = 1 (Underwood)
        # (z-c)^2/a^2 - (y^2+x^2)/b^2 = 1 (Shadow)
        # normal incidence (Underwood->Shadow:  x->z, y^2->y^2+x^2)

        ccc_centered_hyperbola = numpy.array([-1/b_h**2, -1/b_h**2, 1/a_h**2, \
                                              0, 0, 0, \
                                              0, 0, -2*c_h/a_h**2, \
                                              (c_h/a_h)**2 - 1])

        #
        # Ellipse ###########################################################################
        #


        Pe = numpy.sqrt((x_pmin - 2 * c_e) ** 2 + y_pmin ** 2)
        Qe = numpy.sqrt(x_pmin ** 2 + y_pmin ** 2)
        b_e = numpy.sqrt(Pe * Qe) * numpy.sin(self.theta1)
        a_e = numpy.sqrt(c_e ** 2 + b_e ** 2)


        # calculate the intersection x1 of the hyperbola with an ellipse  with c_e and
        # variable eccentricity.
        # then solve numerically x1(ecc) = x_pmin to get the wanted ellipse that intersects
        # the hyperbila at x_pmin

        ecc = (self.calculate_ellipse_eccentricity(c_e, a_h, b_h, c_h, x_pmin))[0]

        a_e = c_e / ecc
        b_e = numpy.sqrt(a_e**2 - c_e**2)

        # Ellipse
        # (x-c_e)^2/a^2 + y^2/b_e^2 = 1 (Underwood)
        # (z-c_e)^2/a^2 + (y^2+x^2)/b_e^2 = 1 (Shadow)
        # normal incidence (Underwood x->z, y->y  ->x)

        ccc_centered_ellipse = numpy.array([1 / b_e ** 2, 1 / b_e ** 2, 1 / a_e ** 2, \
                                            0, 0, 0, \
                                            0, 0, -2 * c_e / a_e ** 2, \
                                            (c_e / a_e) ** 2 - 1])

        # intersection hyperbola ellipse (in fact, redundant because x_he = x_pmin by construction!)

        # hyperbola: (x-c_h)**2/a_h**2 - y**2/b_h**2 = 1
        # ellipse: (x-c_e)**2/a_e**2 + y**2/b_e**2 = 1

        A = 1.0 / a_e ** 2 + (b_h / b_e / a_h) ** 2
        B = -2 * c_e / a_e ** 2 - 2 * c_h * (b_h / b_e / a_h) ** 2
        C = -(b_h / b_e) ** 2 - 1 + (c_e / a_e) ** 2 + (c_h * b_h / b_e / a_h) ** 2

        D = B ** 2 - 4 * A * C
        if D < 0:
            print("\n Cannot calculate ellipse (Delta=%f)...." % D)
        x_he = (-B + numpy.sqrt(D)) / 2 / A
        print("A,B,C,D, x_he: ", A, B, C, D, x_he)
        y_he = b_h * numpy.sqrt(((x_he - c_h) / a_h) ** 2 - 1)
        # y_he2 = b_e * numpy.sqrt(1 - ((x_he - c_e) / a_e) ** 2)

        ############################################################################################################

        if verbose:
            print("f11, f12", f11, f12)
            print("f21, f22", f21, f22)

            print("x_pmin, y_pmin: ", x_pmin, y_pmin)



            print("   theta grazing [deg]: ", theta * 180 / numpy.pi)
            print("   oe 1 can be either parabola or ellipse:")
            print("           Parabola p [m]:", p)
            print("           Ellipse a, b, c [m]: ", a_h, b_h, c_h)
            print("   oe 2 is Hyperbola a, b, c [m]: ", a_h, b_h, c_h)

            print("\n\nCalculated parameters: ")
            print("   ** Origin is at common focus (parabola focus=far hyperbola focus)**")
            print("   Common point (crossing par+hyp): x_pmin, y_pmin: ", x_pmin, y_pmin)
            print("   Common point (crossing ell+hyp): x_he, y_he: ", x_he, y_he)

            print(" ")


            print("\n   normalized ccc_centered_parabola", ccc_centered_parabola / ccc_centered_parabola[0])
            print("   normalized ccc_centered_ellipse", ccc_centered_ellipse / ccc_centered_ellipse[0])
            print("   normalized ccc_centered_hyperbola", ccc_centered_hyperbola / ccc_centered_hyperbola[0])

            print("\n   ccc_centered_parabola", ccc_centered_parabola)
            print("   ccc_centered_ellipse", ccc_centered_ellipse)
            print("   ccc_centered_hyperbola", ccc_centered_hyperbola)



        return  {'ccc1':ccc_centered_parabola,
                 'ccc2':ccc_centered_hyperbola,
                 'ccc3': ccc_centered_ellipse,
                 'p':p,
                 'a_h':a_h, 'b_h':b_h, 'c_h':c_h,
                 'a_e':a_e, 'b_e':b_e, 'c_e':c_e,
                 'x_pmin':x_pmin, 'y_pmin':y_pmin,
                 'x_pmin2': x_pmin1, 'y_pmin2': y_pmin1,
                 'x_he':x_he, 'y_he':y_he,
                 }

    def calculate_ellipse_eccentricity(self, c_e, a_h, b_h, c_h, x_pmin):
        # obtain the eccentricity of the ellipse (given c_e) than crosses the hyperbola (with a_h,b_h,c_h)
        # at abscissa x_pmin

        #     # intersection hyperbola ellipse
        #     # ellipse: (x_pmin-c_e)**2/a_e**2 + y**2/b_e**2 = 1
        #     # hyperbola: (x_pmin-c_h)**2/a_h**2 - y**2/b_h**2 = 1
        #     Method:
        #     i) replace y**2 from hyperbola into ellipse
        #     ii) express a_e and b_e as a function of eccentricity ecc and c_e
        #     iii) change variable delta=(1/ecc**2 -1) thus giving:
        #          a1/(1+delta) + b1 / delta -1 = 0
        #          which is a second-degree equation in delta:
        #          delta**2 + delta * (1-a1-b1) -b1 = 0
        #     iv) Solve the second-degree equation and return ecc
        a1 = ((x_pmin-c_e)/c_e)**2
        b1 = (b_h/c_e)**2 * ( ((x_pmin-c_h)/a_h)**2 -1)
        ecc1 = (0.5 * (1 + a1 + b1 + numpy.sqrt(4 * b1 + (1 - a1 - b1) ** 2))) ** (-1 / 2)
        ecc2 = (0.5 * (1 + a1 + b1 - numpy.sqrt(4 * b1 + (1 - a1 - b1) ** 2))) ** (-1 / 2)
        return ecc1, ecc2

    @classmethod
    def plot_histo(cls, plot_window, x, y, title, xtitle, ytitle, color='blue', replace=True, symbol=''):
        import matplotlib
        matplotlib.rcParams['axes.formatter.useoffset']='False'

        plot_window.addCurve(x, y, title, symbol=symbol, color=color, xlabel=xtitle, ylabel=ytitle, replace=replace) #'+', '^', ','

        if not xtitle is None: plot_window.setGraphXLabel(xtitle)
        if not ytitle is None: plot_window.setGraphYLabel(ytitle)
        if not title is None: plot_window.setGraphTitle(title)

        plot_window.resetZoom()
        plot_window.replot()

        plot_window.setActiveCurve(title)
if __name__ == "__main__":
    app = QApplication(sys.argv)
    w = OWWolterCenteredCalculator()
    w.show()
    app.exec()
    w.saveSettings()



class ConicCoefficientsPreProcessorData:
    def __init__(self,
                conic_coefficient_0=0.0,
                conic_coefficient_1=0.0,
                conic_coefficient_2=0.0,
                conic_coefficient_3=0.0,
                conic_coefficient_4=0.0,
                conic_coefficient_5=0.0,
                conic_coefficient_6=0.0,
                conic_coefficient_7=0.0,
                conic_coefficient_8=0.0,
                conic_coefficient_9=0.0,
                source_plane_distance    = None,
                image_plane_distance     = None,
                angles_respect_to        = None,
                incidence_angle_deg      = None,
                reflection_angle_deg     = None,
                mirror_orientation_angle = None,
                title="",
                 ):
        self.conic_coefficient_0 = conic_coefficient_0
        self.conic_coefficient_1 = conic_coefficient_1
        self.conic_coefficient_2 = conic_coefficient_2
        self.conic_coefficient_3 = conic_coefficient_3
        self.conic_coefficient_4 = conic_coefficient_4
        self.conic_coefficient_5 = conic_coefficient_5
        self.conic_coefficient_6 = conic_coefficient_6
        self.conic_coefficient_7 = conic_coefficient_7
        self.conic_coefficient_8 = conic_coefficient_8
        self.conic_coefficient_9 = conic_coefficient_9


        self.source_plane_distance    = source_plane_distance
        self.image_plane_distance     = image_plane_distance
        self.angles_respect_to        = angles_respect_to
        self.incidence_angle_deg      = incidence_angle_deg
        self.reflection_angle_deg     = reflection_angle_deg
        self.mirror_orientation_angle = mirror_orientation_angle
        self.title = title